#! /usr/bin/env python
"""This program plots several properties related to the interaction of a pair of states s1 and s2.

1. The energies of the two states along the MD trajectory
2. The non-adiabatic coupling between the two states (in meV)
3. The normalized autocorrelation function (NAUF) of a pair of entangled states
 during the MD trajectory.
   It describes how the interaction between two states depend on earlier time.
   The AUF of poorly correlated states goes rapidly to zero.
4. The unnormalized autocorrelation function. It provides similar infos than NAUF,
  but the starting value is average oscillation at t=0 (in eV**2)
5. The spectral density computed from the fourier transform of the NAUF.
 It identifies which phonones coupled to the electronic subsystem.
 In this case, it tells which phonons are involved in the coupling between the
 pair of entangles states.
6. The dephasing function. It indicates how the two entangled states
 decohere over time.
 The inverse of the decoherence rate is related to the homogenoues line
 broadening of emission lines when computed between HOMO and LUMO states.

Note that you have to provide the location of the folder where the NAMD
 hamiltonian elements are stored using the -p flag.
"""

import argparse
import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import interactive

from nanoqm import logger
from nanoqm.analysis import autocorrelate, dephasing, spectral_density
from nanoqm.common import fs_to_nm


def plot_stuff(ens, acf, sd, deph, rate, s1, s2, dt, wsd, wdeph):
    """
    arr - a vector of y-values that are plot
    plot_mean, save_plot - bools telling to plot the mean and save the plot or not, respectively
    """
    dim_x = np.arange(ens.shape[1]) * dt

    plt.figure(1)
    plt.title(f'Energies of state {s1} (red) and {s2} (blue)')
    plt.xlabel('Time (fs)')
    plt.ylabel('Energy (eV)')
    plt.plot(dim_x, ens[0, :], c='r')
    plt.plot(dim_x, ens[1, :], c='b')
    interactive(True)
    plt.show()

    plt.figure(2)
    plt.title(f'Normalized AUF between state {s1} and {s2}')
    plt.xlabel('Time (fs)')
    plt.ylabel('Normalized AUF')
    plt.ylim(-1, 1)
#    plt.plot(ts, acf[:, 1, 0], c='r')
#    plt.plot(ts, acf[:, 1, 1], c='b')
    plt.plot(dim_x, acf[:, 1, 2], c='g')
    plt.axhline(0, c="black")
    interactive(True)
    plt.show()

    plt.figure(3)
    plt.title(f'Un-normalized AUF between state {s1} and {s2}')
    plt.xlabel('Time (fs)')
    plt.ylabel('Un-normalized AUF')
#    plt.plot(ts, acf[:, 1, 0], c='r')
#    plt.plot(ts, acf[:, 1, 1], c='b')
    plt.plot(dim_x, acf[:, 0, 2], c='g')
    plt.axhline(0, c="black")
    interactive(True)
    plt.show()

    plt.figure(4)
    plt.title(f'Dephasing time between state {s1} and {s2}')
    plt.xlabel('Time (fs)')
    plt.ylabel('Dephasing (arbitrary units)')
    plt.xlim(0, wdeph)
    plt.plot(dim_x, deph[:, 0], c='r')
    plt.plot(dim_x, deph[:, 1], c='b')
    logger.info(f'The dephasing time is : {rate:f} fs')
    line_broadening = 1 / rate * fs_to_nm
    logger.info(f'The homogenous line broadening is  : {line_broadening:f} nm')
    interactive(True)
    plt.show()

    plt.figure(5)
    plt.title(f'Influence spectrum state {s1}')
    plt.xlabel('Frequency (cm-1)')
    plt.ylabel('Spectral Density (arbitrary units)')
    plt.xlim(0, wsd)
    plt.plot(sd[0, 1, :], sd[0, 0, :], c='g')
    interactive(True)
    plt.show()

    plt.figure(6)
    plt.title(f'Influence spectrum state {s2}')
    plt.xlabel('Frequency (cm-1)')
    plt.ylabel('Spectral Density (arbitrary units)')
    plt.xlim(0, wsd)
    plt.plot(sd[1, 1, :], sd[1, 0, :], c='g')
    interactive(True)
    plt.show()

    plt.figure(7)
    plt.title(f'Influence spectrum across state {s1} and {s2}')
    plt.xlabel('Frequency (cm-1)')
    plt.ylabel('Spectral Density (arbitrary units)')
    plt.xlim(0, wsd)
    plt.plot(sd[2, 1, :], sd[2, 0, :], c='g')
    interactive(False)
    plt.show()
    fileName = "MOs.png"
    plt.savefig(fileName, format='png', dpi=300)


def main(path_output, s1, s2, dt, wsd, wdeph):
    fn = 'me_energies0'  # it is only necessary the first initial condition
    inpfile = os.path.join(path_output, fn)
    cols = (s1 * 2 + 5, s2 * 2 + 5)
    energies = np.loadtxt(inpfile, usecols=cols)
    # Compute the energy difference between pair of states
    d_E = energies[:, 0] - energies[:, 1]
    # Generate a matrix with s1, s2 and diff between them
    en_states = np.stack((energies[:, 0], energies[:, 1], d_E))
    # Compute autocorrelation function for each column (i.e. state)
    # Take the transpose to have the correct shape for spectral_density
    acf = np.stack(autocorrelate(en_states[i, :])
                   for i in range(en_states.shape[0])).T
    # Compute the spectral density for each column using the normalized acf
    sd = np.stack(spectral_density(acf[:, 1, i], dt)
                  for i in range(en_states.shape[0]))
    # Compute the dephasing time for the uncorrelated acf between two states
    deph, rate = dephasing(acf[:, 0, 2], dt)
    # Plot stuff
    plot_stuff(en_states, acf, sd, deph, rate, s1, s2, dt, wsd, wdeph)


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['p', 's1', 's2', 'dt', 'wsd', 'wdeph']

    return [getattr(args, p) for p in attributes]


# ============<>===============
if __name__ == "__main__":

    msg = "plot_decho -p <path/to/hamiltonians> -s1 <State 1> -s2 <State 2>\
    -dt <time step in fs> \
    -wsd <energy window for spectral density plot in cm-1>\
    -wdeph <time window for dephasing time in fs>"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-p', required=True,
                        help='path to the Hamiltonian files in Pyxaid format')
    parser.add_argument('-s1', required=True, type=int,
                        help='Index of the first state')
    parser.add_argument('-s2', required=True, type=int,
                        help='Index of the second state')
    parser.add_argument('-dt', type=float, default=1.0,
                        help='Index of the second state')
    parser.add_argument('-wsd', type=int, default=1500,
                        help='energy window for spectral density plot in cm-1')
    parser.add_argument('-wdeph', type=int, default=50,
                        help='time window for dephasing time in fs')

    main(*read_cmd_line(parser))



#! /usr/bin/env python
"""
This program plots the average electronic energy during a NAMD simulatons
averaged over several initial conditions.
It plots both the SH and SE population based energies.

Example:

 plot_average_energy.py -p . -nstates 26 -nconds 6

Note that the number of states is the same as given in the pyxaid output.
 It must include the ground state as well.
"""

import numpy as np
import os
import matplotlib.pyplot as plt
import argparse


def plot_stuff(outs, pops):
    """
    energies - a vector of energy values that can be plotted
    """
    dim_x = np.arange(outs.shape[0])

    plot = np.column_stack((outs, pops))
    plt.xlabel('Time (fs)')
    plt.ylabel('Energy (eV)')
    plt.plot(dim_x, plot[:, 0:])

    fileName = "Average_Energy.png"

    plt.show()
    plt.savefig(fileName, format='png', dpi=300)


def read_energies(path, fn, nstates, nconds):
    inpfile = os.path.join(path, fn)
    cols = tuple(range(5, nstates * 2 + 5, 2))
    xs = np.stack(np.loadtxt(f'{inpfile}{j}', usecols=cols)
                  for j in range(nconds)).transpose()
    # Rows = timeframes ; Columns = states ; tensor = initial conditions
    xs = xs.swapaxes(0, 1)
    return xs


def read_pops(path, fn, nstates, nconds):
    inpfile = os.path.join(path, fn)
    cols = tuple(range(3, nstates * 2 + 3, 2))
    xs = np.stack(np.loadtxt(f'{inpfile}{j}', usecols=cols)
                  for j in range(nconds)).transpose()
    # Rows = timeframes ; Columns = states ; tensor = initial conditions
    xs = xs.swapaxes(0, 1)
    return xs


def main(path_output, nstates, nconds):
    outs = read_pops(path_output, 'out', nstates, nconds)
    pops = read_pops(path_output, 'me_pop', nstates, nconds)
    energies = read_energies(path_output, 'me_energies', nstates, nconds)

    # Weighted state energy for a given SH or SH population at time t
    eav_outs = energies * outs
    eav_pops = energies * pops
    # Ensamble average over initial conditions of the electronic energy
    # as a function of time
    el_ene_outs = np.average(np.sum(eav_outs, axis=1), axis=1)
    el_ene_pops = np.average(np.sum(eav_pops, axis=1), axis=1)
    # Ensamble average scaled to the lowest excitation energy.
    # This way the cooling converge to 0.
    lowest_hl_gap = np.average(np.amin(energies[:, 1:, :], axis=1), axis=1)
    ene_outs_ref0 = el_ene_outs - lowest_hl_gap
    ene_pops_ref0 = el_ene_pops - lowest_hl_gap

    plot_stuff(ene_outs_ref0, ene_pops_ref0)


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['p', 'nstates', 'nconds']

    return [getattr(args, p) for p in attributes]


# ============<>===============
if __name__ == "__main__":
    msg = "plot_states_pops -p <path/to/output>\
     -nstates <number of states computed>\
      -nconds <number of initial conditions>"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-p', required=True,
                        help='path to the Hamiltonian files in Pyxaid format')
    parser.add_argument('-nstates', type=int, required=True,
                        help='Number of states')
    parser.add_argument('-nconds', type=int, required=True,
                        help='Number of initial conditions')
    main(*read_cmd_line(parser))



#! /usr/bin/env python
import argparse
from os.path import join

import h5py
from nanoqm import logger


def main(path_hdf5: str, remove_overlaps: bool):
    """Remove unused array from the HDF5."""
    path_swaps = ['swaps']
    paths_overlaps_corrected = [
        join(f'overlaps_{i}/mtx_sji_t0_corrected') for i in range(10000)]
    if remove_overlaps:
        paths_overlaps = [
            join(f'overlaps_{i}/mtx_sji_t0') for i in range(10000)]
    else:
        paths_overlaps = []

    with h5py.File(path_hdf5, 'r+') as f5:
        paths_css = list(filter(lambda x: 'coupling' in x, f5.keys()))
        paths = paths_css + paths_overlaps_corrected + path_swaps + paths_overlaps
        for p in (p for p in paths if p in f5):
            logger.info("removing: ", p)
            del f5[p]


if __name__ == "__main__":
    parser = argparse.ArgumentParser("removeHDF5folders.py")
    parser.add_argument('-hdf5', required=True,
                        help='Path to the HDF5 file')
    parser.add_argument(
        '-o', help='Remove the overlap matrices', action='store_true')
    args = parser.parse_args()
    main(args.hdf5, args.o)



"""Compute multipole integrals using `Libint2 <https://github.com/evaleev/libint/wiki>`.

The interface to the C++ Libint2 library is located at the parent folder,
in the `libint` folder.

Index
-----
.. currentmodule:: nanoqm.integrals.multipole_matrices
.. autosummary::
    get_multipole_matrix
    compute_matrix_multipole

API
---
.. autofunction:: get_multipole_matrix
.. autofunction:: compute_matrix_multipole
"""

from __future__ import annotations

import os
import uuid
from os.path import join
from pathlib import Path
from typing import List, Union, TYPE_CHECKING

import numpy as np
from qmflows.common import AtomXYZ

from .. import logger
from ..common import (DictConfig, is_data_in_hdf5, retrieve_hdf5_data,
                      store_arrays_in_hdf5, tuplesXYZ_to_plams)
from ..compute_integrals import compute_integrals_multipole, get_thread_count, get_thread_type

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from numpy import float64 as f8


def get_multipole_matrix(config: DictConfig, inp: DictConfig, multipole: str) -> NDArray[f8]:
    """Retrieve the `multipole` number `i` from the trajectory. Otherwise compute it.

    Parameters
    ----------
    config
        Global configuration to run a workflow
    inp
        Information about the current point, e.g. molecular geometry.
    multipole
        Either overlap, dipole or quadrupole.

    Returns
    -------
    np.ndarray
        Tensor containing the multipole.

    """
    point = f'point_{inp.i + config.enumerate_from}'
    path_hdf5 = config.path_hdf5
    path_multipole_hdf5 = join(config.orbitals_type, multipole, point)
    matrix_multipole = search_multipole_in_hdf5(
        path_hdf5, path_multipole_hdf5, multipole)

    if matrix_multipole is None:
        matrix_multipole = compute_matrix_multipole(inp.mol, config, multipole)
        store_arrays_in_hdf5(path_hdf5, path_multipole_hdf5, matrix_multipole)

    return matrix_multipole


def search_multipole_in_hdf5(
        path_hdf5: Union[str, Path], path_multipole_hdf5: str, multipole: str) -> None | NDArray[f8]:
    """Search if the multipole is already store in the HDF5."""
    if is_data_in_hdf5(path_hdf5, path_multipole_hdf5):
        logger.info(f"retrieving multipole: {multipole} from the hdf5")
        return retrieve_hdf5_data(path_hdf5, path_multipole_hdf5)

    logger.info(f"computing multipole: {multipole}")
    return None


def compute_matrix_multipole(
        mol: List[AtomXYZ], config: DictConfig, multipole: str) -> NDArray[f8]:
    """Compute a `multipole` matrix: overlap, dipole, etc. for a given geometry `mol`.

    The multipole is Computed in spherical coordinates.

    Note: for the dipole and quadrupole the super_matrix contains all the matrices stack all the
    0-axis.

    Parameters
    ----------
    mol
        Molecule to compute the dipole
    config
        Dictionary with the current configuration
    multipole
        kind of multipole to compute

    Returns
    -------
    np.ndarray
        Matrix with entries <ψi | x^i y^j z^k | ψj>

    """
    path_hdf5 = config.path_hdf5

    # Write molecule in temporal file
    path = join(config.scratch_path, f"molecule_{uuid.uuid4()}.xyz")
    mol_plams = tuplesXYZ_to_plams(mol)
    mol_plams.write(path)

    # name of the basis set
    basis_name = config["cp2k_general_settings"]["basis"]
    thread_count = get_thread_count()
    thread_type = get_thread_type()
    logger.info(f"Will scale over {thread_count} {thread_type} threads")

    if multipole == 'overlap':
        matrix_multipole = compute_integrals_multipole(
            path, path_hdf5, basis_name, multipole)
    elif multipole == 'dipole':
        # The tensor contains the overlap + {x, y, z} dipole matrices
        super_matrix = compute_integrals_multipole(
            path, path_hdf5, basis_name, multipole)
        dim = super_matrix.shape[1]

        # Reshape the super_matrix as a tensor containing overlap + {x, y, z} dipole matrices
        matrix_multipole = super_matrix.reshape(4, dim, dim)

    elif multipole == 'quadrupole':
        # The tensor contains the overlap + {xx, xy, xz, yy, yz, zz} quadrupole matrices
        super_matrix = compute_integrals_multipole(
            path, path_hdf5, basis_name, multipole)
        dim = super_matrix.shape[1]

        # Reshape to 3d tensor containing overlap + {x, y, z} + {xx, xy, xz, yy, yz, zz} quadrupole matrices
        matrix_multipole = super_matrix.reshape(10, dim, dim)

    # Delete the tmp molecule file
    os.remove(path)

    return matrix_multipole



import numpy as np
import numpy.typing as npt

def compute_integrals_couplings(
    __path_xyz_1: str,
    __path_xyz_2: str,
    __path_hdf5: str,
    __basis_name: str,
) -> npt.NDArray[np.float64]: ...

def compute_integrals_multipole(
    __path_xyz_1: str,
    __path_xyz_2: str,
    __basis_name: str,
    __multipole: str,
) -> npt.NDArray[np.float64]: ...

def get_thread_count() -> int: ...

def get_thread_type() -> str: ...



"""Inverse Participation Ratio calculation.

Index
-----
.. currentmodule:: nanoqm.workflows.workflow_ipr
.. autosummary::
    workflow_ipr

"""
__all__ = ['workflow_ipr']

import numpy as np
from scipy.linalg import sqrtm

from qmflows.parsers import readXYZ

from .. import logger
from ..common import (DictConfig, h2ev, number_spherical_functions_per_atom,
                      retrieve_hdf5_data)
from ..integrals.multipole_matrices import compute_matrix_multipole
from .initialization import initialize
from .tools import compute_single_point_eigenvalues_coefficients


def workflow_ipr(config: DictConfig) -> np.ndarray:
    """Compute the Inverse Participation Ratio main function."""
    # Dictionary containing the general information
    config.update(initialize(config))

    # Checking if hdf5 contains the required eigenvalues and coefficientsa
    compute_single_point_eigenvalues_coefficients(config)

    # Logger info
    logger.info("Starting IPR calculation.")

    # Get eigenvalues and coefficients from hdf5
    node_path_coefficients = 'coefficients/point_0/'
    node_path_eigenvalues = 'eigenvalues/point_0'
    atomic_orbitals = retrieve_hdf5_data(config.path_hdf5, node_path_coefficients)
    energies = retrieve_hdf5_data(config.path_hdf5, node_path_eigenvalues)
    energies *= h2ev  # To get them from Hartree to eV

    # Converting the xyz-file to a mol-file
    mol = readXYZ(config.path_traj_xyz)

    # Computing the overlap-matrix S and its square root
    overlap = compute_matrix_multipole(mol, config, 'overlap')
    squared_overlap = sqrtm(overlap)

    # Converting the coeficients from AO-basis to MO-basis
    transformed_orbitals = np.dot(squared_overlap, atomic_orbitals)

    # Now we add up the rows of the c_MO that belong to the same atom
    sphericals = number_spherical_functions_per_atom(
        mol,
        'cp2k',
        config.cp2k_general_settings["basis"],
        config.path_hdf5)  # Array with number of spherical orbitals per atom

    # New matrix with the atoms on the rows and the MOs on the columns
    indices = np.zeros(len(mol), dtype='int')
    indices[1:] = np.cumsum(sphericals[:-1])
    accumulated_transf_orbitals = np.add.reduceat(transformed_orbitals, indices, 0)

    # Finally, we can calculate the IPR
    ipr = np.zeros(accumulated_transf_orbitals.shape[1])

    for i in range(accumulated_transf_orbitals.shape[1]):
        ipr[i] = np.sum(np.absolute(accumulated_transf_orbitals[:, i])**4) / \
            (np.sum(np.absolute(accumulated_transf_orbitals[:, i])**2))**2

    # Lastly, we save the output as a txt-file
    result = np.zeros((accumulated_transf_orbitals.shape[1], 2))
    result[:, 0] = energies
    result[:, 1] = 1.0 / ipr
    np.savetxt('IPR.txt', result)
    return result



"""Module to run restricted and unrestricted calculations."""

from typing import Any, Callable

from noodles import gather
from qmflows import run

from .. import logger
from ..common import DictConfig
from .initialization import initialize


def select_orbitals_type(
        config: DictConfig, workflow: Callable[[DictConfig], Any]) -> Any:
    """Call a workflow using restriced or unrestricted orbitals."""
    # Dictionary containing the general configuration
    config.update(initialize(config))

    if config.orbitals_type != "both":
        logger.info("starting workflow calculation!")
        promises = workflow(config)
        return run(promises, folder=config.workdir, always_cache=False)
    else:
        config_alphas = DictConfig(config.copy())
        config_betas = DictConfig(config.copy())
        config_alphas.orbitals_type = "alphas"
        promises_alphas = workflow(config_alphas)
        config_betas.orbitals_type = "betas"
        promises_betas = workflow(config_betas)
        all_promises = gather(promises_alphas, promises_betas)
        alphas, betas = run(all_promises, folder=config.workdir, always_cache=False)
        return alphas, betas



#! /usr/bin/env python

from os.path import join
import argparse
import h5py
import numpy as np

from nanoqm import logger


def main(project_name, path_hdf5, indices, overlap_flag, mo_flag):

    # Shift indices to start from 0
    indices = np.array(indices) - 1

    # path to the failed points
    if mo_flag:
        mos = [join(project_name, f'point_{i}') for i in indices]
    else:
        mos = []
    if overlap_flag:
        overlaps = [join(project_name, f'overlaps_{i}') for i in indices]
    else:
        overlaps = []

    # Concatenate both Molecular orbitals and Overlaps
    paths = mos + overlaps

    with h5py.File(path_hdf5, 'r+') as f5:
        for p in paths:
            if p in f5:
                logger.info("removing: ", p)
                del f5[p]


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['pn', 'hdf5', 'i', 'o', 'mo']

    return [getattr(args, p) for p in attributes]


if __name__ == "__main__":

    msg = " remove_couplings -pn <ProjectName> -hdf5 <path/to/hdf5 file> -o False"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-pn', required=True,
                        help='project name')
    parser.add_argument('-hdf5', required=True,
                        help='Path to the HDF5')
    parser.add_argument(
        '-o', help='flag to remove the overlaps', action='store_true')
    parser.add_argument('-mo', help='flag to remove the molecular overlaps',
                        action='store_true')
    parser.add_argument('-i', help='Indices of the Molecular orbitals', required=True,
                        nargs='+', type=int)
    main(*read_cmd_line(parser))



"""Schemas to valid user input.

Index
-----
.. currentmodule:: nanoqm.workflows.schemas
.. autosummary::
    {autosummary}

API
---
{autodata}

"""
__all__ = [
    'schema_cp2k_general_settings',
    'schema_derivative_couplings',
    'schema_single_points',
    'schema_distribute_absorption_spectrum',
    'schema_distribute_derivative_couplings',
    'schema_distribute_single_points',
    'schema_absorption_spectrum',
    'schema_ipr',
    'schema_coop']

import os
from numbers import Real
from typing import Any, Dict, Iterable

import pkg_resources as pkg
from schema import And, Optional, Or, Regex, Schema, Use


def equal_lambda(name: str) -> And:
    """Create an schema checking that the keyword matches the expected value."""
    return And(
        str, Use(str.lower), lambda s: s == name)


def any_lambda(array: Iterable[str]) -> And:
    """Create an schema checking that the keyword matches one of the expected values."""
    return And(
        str, Use(str.lower), lambda s: s in array)


def merge(d1: Dict[str, Any], d2: Dict[str, Any]) -> Dict[str, Any]:
    """Merge two dictionaries using without modifying the original."""
    x = d1.copy()

    x.update(d2)

    return x


def _parse_filenames(f: "None | str | list[str]") -> "None | list[str]":
    """Helper function for sanitizing one or multiple filenames."""
    if isinstance(f, str):
        return [f]
    elif f is None or (isinstance(f, list) and all(isinstance(i, str) for i in f)):
        return f
    else:
        raise TypeError(type(f).__name__)


#: Schema to validate the CP2K general settings
schema_cp2k_general_settings = Schema({

    # "Basis set to carry out the quantum chemistry simulation"
    "basis": str,

    # "Pseudo-potential to carry out the quantum chemistry simulation"
    "potential": str,

    # Charge of the system
    Optional("charge", default=0): int,

    # Multiplicity
    Optional("multiplicity", default=1): int,

    # Specify the Cartesian components for the cell vector Units Angstrom
    Optional("cell_parameters", default=10): Or(
        Real,
        lambda xs: len(xs) == 3 and isinstance(xs, list),
        lambda xs: len(xs) == 3 and all(len(r) == 3 for r in xs)),

    # Type of periodicity
    "periodic": any_lambda(("none", "x", "y", "z", "xy", "xy", "yz", "xyz")),

    # Specify the angles between the vectors defining the unit cell
    Optional("cell_angles"): list,

    # Path to the folder containing the basis set specifications
    Optional("path_basis", default=pkg.resource_filename("nanoqm", "basis")): os.path.isdir,

    # Name(s) of the basis set file(s) stored in ``path_basis``
    Optional("basis_file_name", default=None): Use(_parse_filenames),

    # Name(s) of the potential file(s) stored in ``path_basis``
    Optional("potential_file_name", default=None): Use(_parse_filenames),

    # Name(s) of the exchange part of the DFT functional`
    Optional("functional_x", default=None): str,

    # Name(s) of the correlation part of the DFT functional`
    Optional("functional_c", default=None): str,

    # Settings describing the input of the quantum package
    "cp2k_settings_main": object,

    # Settings describing the input of the quantum package
    # to compute the guess wavefunction"
    "cp2k_settings_guess": object,

    # Restart File Name
    Optional("wfn_restart_file_name", default=None): Or(str, None),

    # File containing the Parameters of the cell if those
    # parameters change during the MD simulation.
    Optional("file_cell_parameters", default=None): Or(str, None),

    # Quality of the auxiliar basis cFIT
    Optional("aux_fit", default="verygood"):
        any_lambda(("low", "medium", "good", "verygood", "excellent")),

    # executable name
    # "sdbg" Serial single core testing and debugging
    # "sopt" Serial general single core usage
    # "ssmp" Parallel (only OpenMP), single node, multi core
    # "pdbg" Parallel (only MPI) multi-node testing and debugging
    # "popt" Parallel (only MPI) general usage, no threads
    # "psmp" parallel (MPI + OpenMP) general usage, threading might improve scalability and memory usage

    Optional("executable", default="cp2k.psmp"):
        Regex(r'.*cp2k\.(?:popt|psmp|sdbg|sopt|ssmp|pdbg)', flags=2)  # flag 2 == IGNORECASE
})


#: Dictionary with the options common to all workflows
dict_general_options = {

    # Number of occupied/virtual orbitals to use
    Optional('active_space', default=[10, 10]): And(list, lambda xs: len(xs) == 2),

    # Index of the HOMO
    Optional("nHOMO"): int,

    # Index of the orbitals to compute the couplings
    Optional("mo_index_range"): tuple,

    # "default quantum package used"
    Optional("package_name", default="cp2k"): str,

    # project
    Optional("project_name", default="namd"): str,

    # Working directory
    Optional("scratch_path", default=None): Or(None, str),

    # path to the HDF5 to store the results
    Optional("path_hdf5", default="quantum.hdf5"): str,

    # path to xyz trajectory of the Molecular dynamics
    "path_traj_xyz": os.path.exists,

    # Real from where to start enumerating the folders create for each point
    # in the MD
    Optional("enumerate_from", default=0): int,

    # Ignore the warning issues by the quantum package and keep computing
    Optional("ignore_warnings", default=False): bool,

    # Calculate the guess wave function in either the first point of the
    # trajectory or in all
    Optional("calculate_guesses", default="first"):
    any_lambda(("first", "all")),

    # Units of the molecular geometry on the MD file
    Optional("geometry_units", default="angstrom"):
    any_lambda(("angstrom", "au")),

    # Integration time step used for the MD (femtoseconds)
    Optional("dt", default=1): Real,

    # Deactivate the computation of the orbitals for debugging purposes
    Optional("compute_orbitals", default=True): bool,

    # Flag to remove the log containing the orbitals for debugging purposes
    Optional("remove_log_file", default=False): bool,

    # General settings
    "cp2k_general_settings": schema_cp2k_general_settings,

    # Empty string for restricted calculation or either alpha/beta
    # for unrestricted calculation
    Optional("orbitals_type", default=""): any_lambda(("", "alphas", "betas", "both"))
}

#: Dict with input options to run a derivate coupling workflow
dict_derivative_couplings = {
    # Name of the workflow to run
    "workflow": equal_lambda("derivative_couplings"),

    # Algorithm used to compute the derivative couplings
    Optional("algorithm", default="levine"):
    any_lambda(("levine", "3points")),

    # Use MPI to compute the couplings
    Optional("mpi", default=False): bool,

    # Track the crossing between states
    Optional("tracking", default=True): bool,

    # Write the overlaps in ascii
    Optional("write_overlaps", default=False): bool,

    # Compute the overlap between molecular geometries using a dephase"
    Optional("overlaps_deph", default=False): bool
}

dict_merged_derivative_couplings = merge(
    dict_general_options, dict_derivative_couplings)

#: Schema to validate the input for a derivative coupling calculation
schema_derivative_couplings = Schema(
    dict_merged_derivative_couplings)

#: Schema to validate the input for a job scheduler
schema_job_scheduler = Schema({
    Optional("scheduler", default="slurm"):
    any_lambda(("slurm", "pbs")),
    Optional("nodes", default=1): int,
    Optional("tasks", default=1): int,
    Optional("wall_time", default="01:00:00"): str,
    Optional("job_name", default="namd"): str,
    Optional("queue_name", default="short"): str,
    Optional("load_modules", default=""): str,
    Optional("free_format", default=""): str
})

#: Input options to distribute a job
dict_distribute = {

    Optional("workdir", default=os.getcwd()): str,

    # Number of chunks to split the trajectory
    "blocks": int,

    # Resource manager configuration
    "job_scheduler": schema_job_scheduler,

    # General settings
    "cp2k_general_settings": schema_cp2k_general_settings,


}

#: input to distribute a derivative coupling job
dict_distribute_derivative_couplings = {

    # Name of the workflow to run
    "workflow": equal_lambda("distribute_derivative_couplings")
}


#: Schema to validate the input to distribute a derivate coupling calculation
schema_distribute_derivative_couplings = Schema(
    merge(
        dict_distribute,
        merge(
            dict_merged_derivative_couplings,
            dict_distribute_derivative_couplings)))

#: Input for an absorption spectrum calculation
dict_absorption_spectrum = {

    # Name of the workflow to run
    "workflow": equal_lambda("absorption_spectrum"),

    # Type of TDDFT calculations. Available: sing_orb, stda, stddft
    Optional("tddft", default="stda"): And(
        str, Use(str.lower), lambda s: s in ("sing_orb", "stda", "stdft")),

    # Interval between MD points where the oscillators are computed"
    Optional("stride", default=1): int,

    # description: Exchange-correlation functional used in the DFT
    # calculations,
    Optional("xc_dft", default="pbe"): str
}


dict_merged_absorption_spectrum = merge(
    dict_general_options, dict_absorption_spectrum)

#: Schema to validate the input for an absorption spectrum calculation
schema_absorption_spectrum = Schema(dict_merged_absorption_spectrum)


dict_distribute_absorption_spectrum = {

    # Name of the workflow to run
    "workflow": equal_lambda("distribute_absorption_spectrum")
}

schema_distribute_absorption_spectrum = Schema(
    merge(dict_distribute, merge(
        dict_merged_absorption_spectrum, dict_distribute_absorption_spectrum)))

dict_single_points = {
    # Name of the workflow to run
    "workflow": any_lambda(("single_points", "ipr_calculation", "coop_calculation")),

    # General settings
    "cp2k_general_settings": schema_cp2k_general_settings
}

#: input to distribute single point calculations
dict_distribute_single_points = {

    # Name of the workflow to run
    "workflow": equal_lambda("distribute_single_points")
}

#: Input for a Crystal Orbital Overlap Population calculation
dict_coop = {
    # List of the two elements to calculate the COOP for
    "coop_elements": list}


dict_merged_single_points = merge(dict_general_options, dict_single_points)

#: Schema to validate the input of a single pointe calculation
schema_single_points = Schema(dict_merged_single_points)

#: Schema to validate the input for a Inverse Participation Ratio calculation
schema_ipr = schema_single_points

#: Input for a Crystal Orbital Overlap Population calculation
dict_merged_coop = merge(dict_merged_single_points, dict_coop)

#: Schema to validate the input for a Crystal Orbital Overlap Population calculation
schema_coop = Schema(dict_merged_coop)

#: Schema to validate the input to distribute a single point calculation
schema_distribute_single_points = Schema(
    merge(dict_distribute, merge(
        dict_merged_single_points, dict_distribute_single_points)))



"""Molecular orbitals calculation using CP2K and `QMFlows <https://github.com/SCM-NV/qmflows>`.

API
---
.. autofunction:: calculate_mos

"""

from __future__ import annotations

__all__ = ["calculate_mos", "create_point_folder",
           "split_file_geometries"]

import fnmatch
import os
from collections import defaultdict
from os.path import join
from typing import (Any, DefaultDict, Dict, List, NamedTuple, Sequence, Tuple,
                    Union)

from more_itertools import chunked
from noodles import gather, schedule
from qmflows.common import InfoMO
from qmflows.type_hints import PathLike, PromisedObject
from qmflows.warnings_qmflows import SCF_Convergence_Warning

from .. import logger
from ..common import (DictConfig, Matrix, is_data_in_hdf5,
                      read_cell_parameters_as_array, store_arrays_in_hdf5)
from .scheduleCP2K import prepare_job_cp2k

#: Molecular orbitals from both restricted and unrestricted calculations
OrbitalType = Union[InfoMO, Tuple[InfoMO, InfoMO]]


class JobFiles(NamedTuple):
    """Contains the data to compute the molecular orbitals for a given geometry."""

    get_xyz: PathLike
    get_inp: PathLike
    get_out: PathLike
    get_MO: PathLike


def calculate_mos(config: DictConfig) -> List[str]:
    """Look for the MO in the HDF5 file and compute them if they are not present.

    The orbitals are computed  by splitting the jobs in batches given by
    the ``restart_chunk`` variables. Only the first job is calculated from scratch
    while the rest of the batch uses as guess the wave function of the first calculation
    inthe batch.

    The config dict contains:
        * geometries: list of molecular geometries
        * project_name: Name of the project used as root path for storing data in HDF5.
        * path_hdf5: Path to the HDF5 file that contains the numerical results.
        * folders: path to the directories containing the MO outputs
        * settings_main: Settings for the job to run.
        * calc_new_wf_guess_on_points: Calculate a new Wave function guess in each of the geometries indicated. By Default only an initial guess is computed.
        * enumerate_from: Number from where to start enumerating the folders create for each point in the MD

    Returns
    -------
        paths to the datasets in the HDF5 file containging both the MO energies and MO coefficients

    """
    # Read Cell parameters file
    general = config['cp2k_general_settings']
    file_cell_parameters = general["file_cell_parameters"]
    if file_cell_parameters is not None:
        array_cell_parameters = read_cell_parameters_as_array(file_cell_parameters)[
            1]

    # First calculation has no initial guess
    # calculate the rest of the jobs using the previous point as initial guess
    orbitals = []  # list to the nodes in the HDF5 containing the MOs
    energies = []
    guess_job = None

    # orbital type is either an empty string for restricted calculation
    # or alpha/beta for unrestricted calculations
    orbitals_type = config.orbitals_type

    for j, gs in enumerate(config.geometries):

        # number of the point with respect to all the trajectory
        k = j + config.enumerate_from

        # dictionary containing the information of the j-th job
        dict_input = defaultdict(lambda: None)  # type:  DefaultDict[str, Any]
        dict_input["geometry"] = gs
        dict_input["k"] = k

        # Path where the MOs will be store in the HDF5
        dict_input["node_MOs"] = [
            join(orbitals_type, "eigenvalues", f"point_{k}"),
            join(orbitals_type, "coefficients", f"point_{k}")]

        dict_input["node_energy"] = join(orbitals_type, "energy", f"point_{k}")

        # If the MOs are already store in the HDF5 format return the path
        # to them and skip the calculation
        if config["compute_orbitals"]:
            predicate = dict_input["node_MOs"]
        else:
            predicate = dict_input["node_energy"]

        if is_data_in_hdf5(config.path_hdf5, predicate):
            logger.info(f"point_{k} has already been calculated")
            orbitals.append(dict_input["node_MOs"])
        else:
            logger.info(f"point_{k} has been scheduled")

            # Add cell parameters from file if given
            if file_cell_parameters is not None:
                adjust_cell_parameters(general, array_cell_parameters, j)
            # Path to I/O files
            dict_input["point_dir"] = config.folders[j]
            dict_input["job_files"] = create_file_names(
                dict_input["point_dir"], k)
            dict_input["job_name"] = f'point_{k}'

            # Compute the MOs and return a new guess
            promise_qm = compute_orbitals(config, dict_input, guess_job)

            # Check if the job finishes succesfully
            promise_qm = schedule_check(promise_qm, config, dict_input)

            # Store the computation
            if config["compute_orbitals"]:
                orbitals.append(store_molecular_orbitals(config, dict_input, promise_qm))
            else:
                orbitals.append(None)
            energies.append(store_enery(config, dict_input, promise_qm))

            guess_job = promise_qm

    return gather(gather(*orbitals), gather(*energies))


@schedule
def store_molecular_orbitals(
        config: DictConfig, dict_input: DefaultDict[str, Any], promise_qm: PromisedObject) -> str:
    """Store the MOs in the HDF5.

    Returns
    -------
    str
        Node path in the HDF5

    """
    # Molecular Orbitals
    mos = promise_qm.orbitals

    # Store in the HDF5
    try:
        save_orbitals_in_hdf5(mos, config, dict_input["job_name"])
    # Remove the ascii MO file
    finally:
        if config.remove_log_file:
            work_dir = promise_qm.archive['work_dir']
            path_mos = fnmatch.filter(os.listdir(work_dir), 'mo_*MOLog')[0]
            os.remove(join(work_dir, path_mos))

    return dict_input["node_MOs"]


def save_orbitals_in_hdf5(mos: OrbitalType, config: DictConfig, job_name: str) -> None:
    """Store the orbitals from restricted and unrestricted calculations."""
    if isinstance(mos, InfoMO):
        dump_orbitals_to_hdf5(mos, config, job_name)
    else:
        alphas, betas = mos
        dump_orbitals_to_hdf5(alphas, config, job_name, "alphas")
        dump_orbitals_to_hdf5(betas, config, job_name, "betas")


def dump_orbitals_to_hdf5(
        data: InfoMO, config: DictConfig, job_name: str, orbitals_type: str = "") -> None:
    """Store the result in HDF5 format.

    Parameters
    ----------
    data
        Tuple of energies and coefficients of the molecular orbitals
    config
        Dictionary with the job configuration
    job_name
        Name of the current job
    orbitals_type
        Either an empty string for MO coming from a restricted job or alpha/beta
        for unrestricted MO calculation
    """
    for name, array in zip(("eigenvalues", "coefficients"), (data.eigenvalues, data.eigenvectors)):
        path_property = join(orbitals_type, name, job_name)
        store_arrays_in_hdf5(config.path_hdf5, path_property, array)


@schedule
def store_enery(
        config: DictConfig, dict_input: Dict, promise_qm: PromisedObject) -> str:
    """Store the total energy in the HDF5 file.

    Returns
    -------
    str
        Node path to the energy in the HDF5

    """
    store_arrays_in_hdf5(
        config.path_hdf5, dict_input['node_energy'], promise_qm.energy)

    logger.info(
        f"Total energy of point {dict_input['k']} is: {promise_qm.energy}")

    return dict_input["node_energy"]


def compute_orbitals(
        config: DictConfig, dict_input: Dict, guess_job: PromisedObject) -> PromisedObject:
    """Call a Quantum chemisty package to compute the MOs.

    When finish store the MOs in the HdF5 and returns a new guess.
    """
    dict_input["job_files"] = create_file_names(
        dict_input["point_dir"], dict_input["k"])

    # Calculating initial guess
    compute_guess = config.calc_new_wf_guess_on_points is not None

    # A job  is a restart if guess_job is None and the list of
    # wf guesses are not empty
    is_restart = guess_job is None and compute_guess

    pred = (dict_input['k']
            in config.calc_new_wf_guess_on_points) or is_restart

    general = config.cp2k_general_settings

    if pred:
        guess_job = prepare_job_cp2k(
            general["cp2k_settings_guess"], dict_input, guess_job)

    promise_qm = prepare_job_cp2k(
        general["cp2k_settings_main"], dict_input, guess_job)

    return promise_qm


@schedule
def schedule_check(
        promise_qm: PromisedObject, config: DictConfig, dict_input: DictConfig) -> PromisedObject:
    """Check wether a calculation finishes succesfully otherwise run a new guess."""
    job_name = dict_input["job_name"]
    point_dir = dict_input["point_dir"]

    # Warnings of the computation
    warnings = promise_qm.warnings

    # Check for SCF convergence errors
    if not config.ignore_warnings and warnings is not None and any(
            w == SCF_Convergence_Warning for msg, w in warnings.items()):
        # Report the failure
        msg = f"Job: {job_name} Finished with Warnings: {warnings}"
        logger.warning(msg)

        # recompute a new guess
        msg1 = "Computing a new wave function guess for job: {job_name}"
        logger.warning(msg1)

        # Remove the previous ascii file containing the MOs
        msg2 = f"removing file containig the previous failed MOs of {job_name}"
        logger.warning(msg2)
        path = fnmatch.filter(os.listdir(point_dir), 'mo*MOLog')[0]
        os.remove(join(point_dir, path))

        # Compute new guess at point k
        config.calc_new_wf_guess_on_points.append(dict_input["k"])
        return compute_orbitals(config, dict_input, None)
    else:
        return promise_qm


def create_point_folder(
        work_dir: str | os.PathLike[str], n: int, enumerate_from: int) -> List[str]:
    """Create a new folder for each point in the MD trajectory."""
    folders = []
    for k in range(enumerate_from, n + enumerate_from):
        new_dir = join(work_dir, f'point_{k}')
        os.makedirs(new_dir, exist_ok=True)
        folders.append(new_dir)

    return folders


def split_file_geometries(path_xyz: PathLike) -> Sequence[str]:
    """Read a set of molecular geometries in xyz format."""
    # Read Cartesian Coordinates
    with open(path_xyz) as f:
        xss = iter(f.readlines())

    data = []
    while True:
        try:
            natoms = int(next(xss).split()[0])
            molecule = "".join([next(xss) for _ in range(natoms + 1)])
            data.append(f"{natoms}\n{molecule}")
        except StopIteration:
            break

    return data


def create_file_names(work_dir: str | os.PathLike[str], i: int) -> JobFiles:
    """Create a namedTuple with the name of the 4 files used for each point in the trajectory."""
    file_xyz = join(work_dir, f'coordinates_{i}.xyz')
    file_inp = join(work_dir, f'point_{i}.inp')
    file_out = join(work_dir, f'point_{i}.out')
    file_mo = join(work_dir, f'mo_coeff_{i}.out')

    return JobFiles(file_xyz, file_inp, file_out, file_mo)


def adjust_cell_parameters(
        general: DictConfig,
        array_cell_parameters: Matrix,
        j: int) -> None:
    """Adjust the cell parameters on the fly.

    If the cell parameters change during the MD simulations, adjust them
    for the molecular orbitals computation.
    """
    for s in (
        general[p] for p in (
            'cp2k_settings_main',
            'cp2k_settings_guess')):
        s.cell_parameters = array_cell_parameters[j, 2:11].reshape(
            3, 3).tolist()
        s.cell_angles = None



"""Tools for postprocessing."""
from .tools import (autocorrelate, convolute, dephasing, func_conv,
                    gauss_function, parse_list_of_lists, read_couplings,
                    read_energies, read_energies_pyxaid, read_pops_pyxaid,
                    spectral_density)

__all__ = [
    'autocorrelate', 'dephasing', 'convolute', 'func_conv', 'gauss_function',
    'parse_list_of_lists', 'read_couplings', 'read_energies',
    'read_energies_pyxaid', 'read_pops_pyxaid', 'spectral_density']



"""The Nano-QMFlows version tuple."""

from nanoutils import VersionInfo
from packaging.version import Version

from ._version import __version__

__all__ = ["version_info"]

VERSION = Version(__version__)
version_info = VersionInfo._make(VERSION.release[:3])



#!/usr/bin/env python
import argparse

import matplotlib.pyplot as plt

from nanoqm.analysis import rdf


def main(fn, atoms_i, atoms_j, dr, rmax):
    x, y = rdf(fn, atoms_i, atoms_j, dr, rmax)
    plt.xlabel('radius R (Angstrom)')
    plt.ylabel('Pair distributon function (arb. un.)')
    plt.plot(x, y)
    plt.show()


def read_cmd_line(parser):
    """Parse Command line options."""
    args = parser.parse_args()

    attributes = ['file', 'atom1', 'atom2', 'bin', 'rmax']

    return [getattr(args, p) for p in attributes]


if __name__ == "__main__":
    msg = "rdf -file <path/to/filename> -atom1 <atom typer> -atom2 <atom type> -bin <size of the histogram bin> -rmax <cutoff>"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument(
        '-file', required=True, help='name of the xyz trajectory file')
    parser.add_argument(
        '-atom1', required=True, help='first atom type to compute rdf')
    parser.add_argument(
        '-atom2', required=True, help='second atom type to compute rdf')
    parser.add_argument(
        '-bin', required=False, type=float, default=0.05, help='size of the bin')
    parser.add_argument(
        '-rmax', required=False, type=float, default=12.0, help='cutoff in Angstrom for computing the rdf')
    main(*read_cmd_line(parser))



"""Workflow to compute the derivate coupling between states.

The ``workflow_derivative_couplings`` expected a file with a trajectory-like
file with the molecular geometries to compute the couplings.

Index
-----
.. currentmodule:: nanoqm.workflows.workflow_coupling
.. autosummary::

"""

from __future__ import annotations

__all__ = ['workflow_derivative_couplings']

import os
from os.path import join
from typing import List, Tuple, Union

from noodles import gather, schedule, unpack
from noodles.interface import PromisedObject
from qmflows.type_hints import PathLike

from .. import logger
from ..common import DictConfig
from ..schedule.components import calculate_mos
from ..schedule.scheduleCoupling import (calculate_overlap, lazy_couplings,
                                         write_hamiltonians)
from .orbitals_type import select_orbitals_type

#: Type defining the derivative couplings calculation
ResultPaths = Tuple[List[str], List[str]]


def workflow_derivative_couplings(
        config: DictConfig) -> Union[ResultPaths, Tuple[ResultPaths, ResultPaths]]:
    """Compute the derivative couplings for a molecular dynamic trajectory.

    Parameters
    ----------
    config
        Dictionary with the configuration to run the workflows

    Return
    ------
    Folders where the Hamiltonians are stored.

    """
    return select_orbitals_type(config, run_workflow_couplings)


def run_workflow_couplings(config: DictConfig) -> PromisedObject:
    """Run the derivative coupling workflow using `config`."""
    # compute the molecular orbitals
    logger.info("starting couplings calculation!")
    mo_paths_hdf5, energy_paths_hdf5 = unpack(calculate_mos(config), 2)

    # Overlap matrix at two different times
    promised_overlaps = calculate_overlap(config, mo_paths_hdf5)

    # Calculate Non-Adiabatic Coupling
    promised_crossing_and_couplings = lazy_couplings(config, promised_overlaps)

    # Write the results in PYXAID format
    config.path_hamiltonians = create_path_hamiltonians(config.workdir, config.orbitals_type)

    # Inplace scheduling of write_hamiltonians function.
    # Equivalent to add @schedule on top of the function
    schedule_write_ham = schedule(write_hamiltonians)

    # Number of matrix computed
    config["npoints"] = len(config.geometries) - 2

    # Write Hamilotians in PYXAID format
    promise_files = schedule_write_ham(
        config, promised_crossing_and_couplings, mo_paths_hdf5)

    return gather(promise_files, energy_paths_hdf5)


def create_path_hamiltonians(workdir: str | os.PathLike[str], orbitals_type: str) -> str:
    """Create the Paths to store the resulting hamiltonians."""
    prefix = "hamiltonians"
    name = prefix if not orbitals_type else f"{orbitals_type}_{prefix}"
    path_hamiltonians = join(workdir, name)
    if not os.path.exists(path_hamiltonians):
        os.makedirs(path_hamiltonians)

    return path_hamiltonians



"""Workflow to perform single point calculation in a trajectory.

Index
-----
.. currentmodule:: nanoqm.workflows.workflow_single_points
.. autosummary::
    workflow_single_points

"""

__all__ = ['workflow_single_points']


from typing import List

from qmflows import run
from qmflows.packages import Result

from .. import logger
from ..common import DictConfig
from ..schedule.components import calculate_mos
from .initialization import initialize


def workflow_single_points(config: DictConfig) -> Result:
    """Perform single point calculations for a given trajectory.

    Parameters
    ----------
    config
        Input to run the workflow.

    Returns
    -------
    List with the node path to the molecular orbitals in the HDF5.

    """
    # Dictionary containing the general configuration
    config.update(initialize(config))

    logger.info("starting!")

    # compute the molecular orbitals
    # Unpack
    mo_paths_hdf5 = calculate_mos(config)
    # Pack
    results = run(mo_paths_hdf5, folder=config.workdir)

    return results



"""Compute couplings and molecular orbitals.

Use the `Noodles schedule method <http://nlesc.github.io/noodles>` to generate
a dependency graph for running the jobs.

Index
-----
.. currentmodule:: nanoqm.schedule.scheduleCoupling
.. autosummary::
    calculate_overlap
    lazy_couplings
    write_hamiltonians

API
---
.. autofunction:: calculate_overlap
.. autofunction:: lazy_couplings
.. autofunction:: write_hamiltonians

"""

from __future__ import annotations

import os
from os.path import join
from typing import List, Tuple, TYPE_CHECKING

import numpy as np
from noodles import schedule
from scipy.optimize import linear_sum_assignment
from qmflows.parsers import parse_string_xyz

from .. import logger
from ..common import (DictConfig, Matrix, MolXYZ, Tensor3D, Vector, hbar,
                      femtosec2au, h2ev, is_data_in_hdf5, retrieve_hdf5_data,
                      store_arrays_in_hdf5)
from ..integrals import (calculate_couplings_3points,
                         calculate_couplings_levine,
                         compute_overlaps_for_coupling, correct_phases)
from ..integrals.nonAdiabaticCoupling import (compute_range_orbitals,
                                              read_overlap_data)

if TYPE_CHECKING:
    import numpy.typing as npt

__all__ = ["calculate_overlap", "lazy_couplings", "write_hamiltonians"]


@schedule
def lazy_couplings(
    config: DictConfig,
    paths_overlaps: List[str],
) -> Tuple[npt.NDArray[np.int_], List[str]]:
    """Compute the Nonadibatic coupling.

    The coupling is computed sing a either 3-point approximation. See:
    The Journal of Chemical Physics 137, 22A514 (2012); doi: 10.1063/1.4738960

    or a 2Point approximation using an smoothing function:
    J. Phys. Chem. Lett. 2014, 5, 2351−2356; doi: 10.1021/jz5009449

    Notice that the states can cross frequently due to unavoided crossing and
    such crossing must be track. see:
    J. Chem. Phys. 137, 014512 (2012); doi: 10.1063/1.4732536

    Parameters
    ----------
    config:
        Configuration of the current task
    paths_overlaps:
        path to the HDF5 node where the overlaps are stored.

    Returns
    -------
    Tuple
        with the swaps and couplings in each point.

    """
    if config.tracking:
        fixed_phase_overlaps, swaps = compute_the_fixed_phase_overlaps(
            paths_overlaps, config)
    else:
        # Do not track the crossings
        mtx_0 = retrieve_hdf5_data(config.path_hdf5, paths_overlaps[0])
        _, dim = mtx_0.shape
        overlaps = np.stack(
            retrieve_hdf5_data(config.path_hdf5, paths_overlaps))
        nOverlaps, nOrbitals, _ = overlaps.shape
        swaps = np.tile(np.arange(nOrbitals), (nOverlaps + 1, 1))
        mtx_phases = compute_phases(overlaps, nOverlaps, dim)
        fixed_phase_overlaps = correct_phases(overlaps, mtx_phases)

    # Write the overlaps in text format
    if config.write_overlaps:
        logger.debug("Writing down the overlaps in ascii format")
        write_overlaps_in_ascii(fixed_phase_overlaps)

    # Compute the couplings using either the levine method
    # or the 3Points approximation
    coupling_algorithms = {'levine': (calculate_couplings_levine, 1),
                           '3points': (calculate_couplings_3points, 2)}
    # Choose an algorithm to compute the couplings
    fun_coupling, step = coupling_algorithms[config["algorithm"]]
    config["fun_coupling"] = fun_coupling

    # Number of couplings to compute
    nCouplings = fixed_phase_overlaps.shape[0] - step + 1
    couplings = [calculate_couplings(config, i, fixed_phase_overlaps)
                 for i in range(nCouplings)]

    return swaps, couplings


def compute_the_fixed_phase_overlaps(
    paths_overlaps: List[str],
    config: DictConfig,
) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.int_]]:
    """Fix the phase of the overlaps.

    First track the unavoided crossings between Molecular orbitals and
    finally correct the phase for the whole trajectory.
    """
    number_of_frames = len(paths_overlaps)
    # Pasth to the overlap matrices after the tracking
    # and phase correction
    roots = [join(config.orbitals_type, f'overlaps_{i}')
             for i in range(config.enumerate_from, number_of_frames + config.enumerate_from)]
    paths_corrected_overlaps = [join(r, 'mtx_sji_t0_corrected') for r in roots]
    # Paths inside the HDF5 to the array containing the tracking of the
    # unavoided crossings
    path_swaps = join(config.orbitals_type, 'swaps')

    # Compute the corrected overlaps if not avaialable in the HDF5
    all_data_in_hdf5 = all(is_data_in_hdf5(config.path_hdf5, path_data)
                           for path_data in (paths_corrected_overlaps[0], path_swaps))
    if not all_data_in_hdf5:
        # Compute the dimension of the coupling matrix
        mtx_0 = retrieve_hdf5_data(config.path_hdf5, paths_overlaps[0])
        _, dim = mtx_0.shape

        # Read all the Overlaps
        overlaps = np.stack([retrieve_hdf5_data(config.path_hdf5, ps)
                             for ps in paths_overlaps])

        # Number of couplings to compute
        nCouplings = overlaps.shape[0]

        # Compute the unavoided crossing using the Overlap matrix
        # and correct the swaps between Molecular Orbitals
        logger.debug("Computing the Unavoided crossings, "
                     "Tracking the crossings between MOs")
        overlaps, swaps = track_unavoided_crossings(overlaps, config.nHOMO)

        # Compute all the phases taking into account the unavoided crossings
        logger.debug("Computing the phases of the MOs")
        mtx_phases = compute_phases(overlaps, nCouplings, dim)

        # Fixed the phases of the whole set of overlap matrices
        fixed_phase_overlaps = correct_phases(overlaps, mtx_phases)

        # Store corrected overlaps in the HDF5
        store_arrays_in_hdf5(config.path_hdf5, paths_corrected_overlaps,
                             fixed_phase_overlaps)

        # Store the Swaps tracking the crossing
        store_arrays_in_hdf5(config.path_hdf5, path_swaps, swaps, dtype=np.int32)
    else:
        # Read the corrected overlaps and the swaps from the HDF5
        fixed_phase_overlaps = np.stack(
            retrieve_hdf5_data(config.path_hdf5, paths_corrected_overlaps))
        swaps = retrieve_hdf5_data(config.path_hdf5, path_swaps)

    return fixed_phase_overlaps, swaps


def calculate_couplings(config: DictConfig, i: int, fixed_phase_overlaps: Tensor3D) -> str:
    """Compute couplings for the i-th geometry.

    Search for the ith Coupling in the HDF5, if it is not available compute it
    using the 3 points approximation and store it in the HDF5.

    Returns
    -------
    str
        Path to the couplings store in the HDF5 file.

    """
    # time in atomic units
    dt_au = config.dt * femtosec2au

    # Path were the couplinp is store
    k = i + config.enumerate_from
    path = join(config.orbitals_type, f'coupling_{k}')

    # Skip the computation if the coupling is already done
    if is_data_in_hdf5(config.path_hdf5, path):
        logger.info(f"Coupling: {path} has already been calculated")
        return path
    else:
        logger.info(f"Computing coupling: {path}")
        if config.algorithm == 'levine':
            # Extract the overlap matrices involved in the coupling computation
            sji_t0 = fixed_phase_overlaps[i]
            # Compute the couplings with the phase corrected overlaps
            couplings = config["fun_coupling"](
                dt_au, sji_t0, sji_t0.transpose())
        elif config.algorithm == '3points':
            sji_t0, sji_t1 = fixed_phase_overlaps[i: i + 2]
            couplings = config["fun_coupling"](dt_au, sji_t0, sji_t0.transpose(), sji_t1,
                                               sji_t1.transpose())

            # Store the Coupling in the HDF5
        store_arrays_in_hdf5(config.path_hdf5, path, couplings)

        return path


def compute_phases(overlaps: Tensor3D, nCouplings: int,
                   dim: int) -> Matrix:
    """Compute the phase of the state_i at time t + dt.

    It uses the following equation:
    .. python::
        phase_i(t+dt) = Sii(t) * phase_i(t)
    """
    # initial references
    references = np.ones(dim)

    # Matrix containing the phases
    mtx_phases = np.empty((nCouplings + 1, dim))
    mtx_phases[0, :] = references

    # Compute the phases at times t + dt using the phases at time t
    for i in range(nCouplings):
        Sji_t = overlaps[i].reshape(dim, dim)

        # Compute the phase at time t
        phases = np.sign(np.diag(Sji_t)) * references
        mtx_phases[i + 1] = phases
        references = phases

    # Print phases (debug)
    logger.info(f"mtx_phases are stored in HDF5 node: {mtx_phases}")

    return mtx_phases


def track_unavoided_crossings(
    overlaps: npt.NDArray[np.float64],
    nHOMO: int,
) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.int_]]:
    """Track the index of the states if there is a crossing.

    It uses the algorithm  described at:
    J. Chem. Phys. 137, 014512 (2012); doi: 10.1063/1.4732536.

    Returns
    -------
    tuple
        containing the corrected overlaps and the crossings

    """
    # 3D array containing the costs
    # Notice that the cost is compute on half of the overlap matrices
    # correspoding to Sji_t, the other half corresponds to Sij_t
    nOverlaps, nOrbitals, _ = overlaps.shape

    # Indexes taking into account the crossing
    # There are 2 Overlap matrices at each time t
    indexes = np.empty((nOverlaps + 1, nOrbitals), dtype=np.int_)
    indexes[0] = np.arange(nOrbitals, dtype=np.int_)

    # Track the crossing using the overlap matrices

    for k in range(nOverlaps):
        # Cost matrix to track the corssings
        logger.info(f"Tracking crossings at time: {k}")
        cost_mtx_homos = np.negative(overlaps[k, :nHOMO, :nHOMO] ** 2)
        cost_mtx_lumos = np.negative(overlaps[k, nHOMO:, nHOMO:] ** 2)

        # Compute the swap at time t + dt using two set of Orbitals:
        # HOMOs and LUMOS
        swaps_homos = linear_sum_assignment(cost_mtx_homos)[1]
        swaps_lumos = linear_sum_assignment(cost_mtx_lumos)[1]
        total_swaps = np.concatenate((swaps_homos, swaps_lumos + nHOMO))
        indexes[k + 1] = total_swaps

        # update the overlaps at times > t with the previous swaps
        if k != (nOverlaps - 1):  # last element
            k1 = k + 1
            # Update the matrix Sji at time t
            overlaps[k] = swap_columns(overlaps[k], total_swaps)
            # Update all the matrices Sji at time > t
            overlaps[k1:] = swap_forward(overlaps[k1:], total_swaps)
    # Accumulate the swaps
    acc = indexes[0]
    arr = np.empty(indexes.shape, dtype=np.int_)
    arr[0] = acc

    # Fold accumulating the crossings
    for i in range(nOverlaps):
        acc = acc[indexes[i + 1]]
        arr[i + 1] = acc

    return overlaps, arr


def swap_forward(
    overlaps: npt.NDArray[np.float64],
    swaps: npt.NDArray[np.integer],
) -> npt.NDArray[np.float64]:
    """Track all the crossings that happend previous to the current time.

    Swap the index i corresponding to the ith Molecular orbital
    with the corresponding swap at time t0.
    Repeat the same procedure with the index and swap at time t1.
    The swaps are compute with the linear sum assignment algorithm from Scipy.
    https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.optimize.linear_sum_assignment.html

    """
    for i, mtx in enumerate(np.rollaxis(overlaps, 0)):
        overlaps[i] = mtx[:, swaps][swaps]

    return overlaps


@schedule
def calculate_overlap(config: DictConfig, mo_paths_hdf5: List[str]) -> List[str]:
    """Calculate the Overlap matrices.

    Parameters
    ----------
    config
        Configuration of the current job
    mo_paths_hdf5
        Node paths to the molecular orbitals in the HDF5

    Returns
    -------
        Node paths to the overlaps stored in the HDF5
    """
    # Number of couplings to compute
    npoints = len(config.geometries) - 1
    # Check what are the missing Couplings
    all_overlaps_paths = [create_overlap_path(
        config, i) for i in range(npoints)]
    overlap_is_done = [check_if_overlap_is_done(
        config, p) for p in all_overlaps_paths]

    paths = []
    for i in range(npoints):
        if overlap_is_done[i]:
            p = all_overlaps_paths[i]
        else:
            p = single_machine_overlaps(config, mo_paths_hdf5, i)
        paths.append(p)

    return paths


def single_machine_overlaps(
        config: DictConfig, mo_paths_hdf5: List[str], i: int) -> str:
    """Compute the overlaps in the CPUs avaialable on the local machine.

    Returns
    -------
    str
        Node path to the overlaps store in the HDF5

    """
    # Data to compute the overlaps
    pair_molecules = select_molecules(config, i)
    mo_paths = [mo_paths_hdf5[i + j][1] for j in range(2)]
    coefficients = read_overlap_data(config, mo_paths)

    # Compute the overlap
    overlaps = compute_overlaps_for_coupling(
        config, pair_molecules, coefficients)

    logger.info(f"overlap for point {i} was sucessfully computed!")

    # Store the array in the HDF5
    overlaps_paths_hdf5 = create_overlap_path(config, i)
    store_arrays_in_hdf5(config.path_hdf5, overlaps_paths_hdf5, overlaps)

    return overlaps_paths_hdf5


def create_overlap_path(config: DictConfig, i: int) -> str:
    """Create the path inside the HDF5 where the overlap is going to be store."""
    root = join(config.orbitals_type, 'overlaps_{}'.format(
        i + config.enumerate_from))
    return join(root, 'mtx_sji_t0')


def select_molecules(config: DictConfig, i: int) -> Tuple[MolXYZ, MolXYZ]:
    """Select the pairs of molecules to compute the couplings."""
    k = 0 if config.overlaps_deph else i
    return (
        parse_string_xyz(config.geometries[k]),
        parse_string_xyz(config.geometries[i + 1]),
    )


def check_if_overlap_is_done(config: DictConfig, overlaps_paths_hdf5: str) -> bool:
    """Search for a given Overlap inside the HDF5."""
    if is_data_in_hdf5(config.path_hdf5, overlaps_paths_hdf5):
        logger.info(f"{overlaps_paths_hdf5} Overlaps are already in the HDF5")
        return True
    else:
        logger.info(f"Computing: {overlaps_paths_hdf5}")
        return False


def write_hamiltonians(
        config: DictConfig,
        crossing_and_couplings: Tuple[np.ndarray, List[Matrix]],
        mo_paths_hdf5: List[str]) -> List[Tuple[str, str]]:
    """Write the real and imaginary components of the hamiltonian.

    It uses both the orbitals energies and the derivative coupling accoring to:
    http://pubs.acs.org/doi/abs/10.1021/ct400641n
    .. Note::
        **Units are: Rydbergs**.

    Parameters
    ----------
    config
        Configuration of the current task
    crossing_and-couplings
        Tuple of the states' crossings and couplings
    mo_path_hdf5
        Node path to the molecular orbitals in the HDF5

    Returns
    -------
    tuple
        Files containing the Hamiltonian Real and imaginary components

    """
    swaps, path_couplings = crossing_and_couplings
    nHOMO = config.nHOMO
    mo_index_range = config.mo_index_range

    def write_pyxaid_format(arr, fileName):
        np.savetxt(fileName, arr, fmt='%10.5e', delimiter='  ')

    def write_data(i):
        j = i + config.enumerate_from
        path_coupling = path_couplings[i]
        css = retrieve_hdf5_data(config.path_hdf5, path_coupling)

        # Extract the energy values at time t
        # The first coupling is compute at time t + dt
        # Then I'm shifting the energies dt to get the correct value
        energies_t0 = retrieve_hdf5_data(config.path_hdf5, mo_paths_hdf5[i][0])
        energies_t1 = retrieve_hdf5_data(
            config.path_hdf5, mo_paths_hdf5[i + 1][0])

        # Return the average between time t and t + dt
        energies = np.average((energies_t0, energies_t1), axis=0)

        # Print Energies in the range given by the user
        if all(x is not None for x in [nHOMO, mo_index_range]):
            lowest, highest = compute_range_orbitals(config)
            energies = energies[lowest: highest]

        # Swap the energies of the states that are crossing
        energies = energies[swaps[i + 1]]

        # FileNames
        path_dir_results = config.path_hamiltonians
        file_ham_im = join(path_dir_results, f'Ham_{i}_im')
        file_ham_re = join(path_dir_results, f'Ham_{j}_re')

        # Time units are atomic units. Convert them in fs-1, then in eV by hbar (eV * fs)
        ham_im = css * femtosec2au * hbar

        # Set the diagonal of the imaginary matrix to 0
        np.fill_diagonal(ham_im, 0)

        # Energies in eV
        ham_re = np.diag(h2ev * energies)

        write_pyxaid_format(ham_im, file_ham_im)
        write_pyxaid_format(ham_re, file_ham_re)

        return (file_ham_im, file_ham_re)

    # The couplings are compute at time t + dt therefore
    # we associate the energies at time t + dt with the corresponding coupling
    return [write_data(i) for i in range(config.npoints)]


def swap_columns(arr: Matrix, swaps_t: Vector) -> Matrix:
    """Swap only columns at t + dt to reconstruct the original overlap matrix."""
    return np.transpose(np.transpose(arr)[swaps_t])


def write_overlaps_in_ascii(overlaps: Tensor3D) -> None:
    """Write the corrected overlaps in text files."""
    if not os.path.isdir('overlaps'):
        os.mkdir('overlaps')

    # write overlaps
    nframes = overlaps.shape[0]
    for k in range(nframes):
        mtx_Sji = overlaps[k]
        path_Sji = f'overlaps/mtx_Sji_{k}'

        np.savetxt(path_Sji, mtx_Sji, fmt='%10.5e', delimiter='  ')



#! /usr/bin/env python
"""
This script plots the stacked contribution of each atom type to a given MO.
You can also group together the contribution of a set of atoms
(for example that form a ligand). Usage:

plot_dos.py -ligand 2 3 -emin -5.0 -emax 2.0

When you use the ligand flag you imply that some atom kinds,
in this case k2 and k3, can be grouped together.
If you dont use the ligand flag, each atom contributes independently.
emin and emax indicates the energy window to make the plot.
"""


import argparse
import glob

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import interactive

from nanoqm import logger


def readatom(filename):
    # In the first line in column 6, the atom is defined
    with open(filename, 'r') as f:
        atom = f.readline().split()[6]
    return atom


def g(x_real, x_grid, delta):
    return np.exp(-2 * (x_grid - x_real) ** 2 / delta ** 2)


def convolute(x, y, x_points, sigma):
    # Compute gaussian prefactor
    prefactor = np.sqrt(2.0) / (sigma * np.sqrt(np.pi))
    # Convolute spectrum over grid
    y_points = prefactor * np.stack(
        np.sum(y * g(x, x_point, sigma)) for x_point in x_points)
    return y_points


def plot_stuff(ys, energies, legends, emin, emax):
    # In case you need more colors, this list should be expanded
    colors = ['black', 'orange', 'blue', 'red',
              'green', 'magenta', 'cyan', 'black', 'yellow']
    #   First plot
    left = 0
    for i in range(ys.shape[1]):
        plt.barh(energies, ys[:, i], height=0.02, left=left,
                 lw=0, color=colors[i], label=legends[i])
        left += ys[:, i]

    plt.figure(1)
    plt.ylim(emin, emax)
    plt.xlim(0, 1)
    plt.tick_params(direction='out')
#    plt.legend(loc='upper right').draggable()
    interactive(True)
    plt.show()

#   Prepare the cumulative density of states
    sigma = 0.2
    x_points = np.linspace(energies[0], energies[-1], 1000)
    # Take transpose to have it in (x_points, # dos) format
    dos = np.stack(convolute(
        energies, ys[:, i], x_points, sigma) for i in range(ys.shape[1])).T
    # Cumulate dos for all atoms/ligands
    cum_dos = np.stack(np.sum(dos[:, 0:i+1], axis=1)
                       for i in range(dos.shape[1])).T
    #   Plotting now
    plt.figure(2)
    ref = np.zeros(x_points.size)
    # Add a zero line for filling colors
    cum_dos = np.column_stack((ref, cum_dos))
    plt.xlim(emin, emax)
    max_y_plot = np.amax(
        cum_dos[np.where(x_points > emin) and np.where(x_points < emax), -1])
    plt.ylim(0, max_y_plot)
    for i in range(cum_dos.shape[1]-1):
        plt.fill_between(
            x_points, cum_dos[:, i], cum_dos[:, i + 1], color=colors[i], label=legends[i])
    plt.legend(loc='upper right').draggable()
    plt.tick_params(direction='out')
    interactive(False)
    plt.show()


def main(group, emin, emax):
    # Check files with PDOS
    files = sorted(glob.glob('*-k*.pdos'))
    # Define the atom type for each DOS file
    legends = [readatom(files[i]) for i in range(len(files))]
    logger.info(files, legends)
    # MO energies
    energies = np.loadtxt(files[0], usecols=1)
    # Convert energies to eV
    energies *= 27.211
    # Occupation
    occ = np.loadtxt(files[0], usecols=2)
    lumos_indx = np.where(occ == 0)
    lumo_indx = lumos_indx[0][0]
    homo_indx = lumo_indx - 1
    hl_gap = (energies[lumo_indx] - energies[homo_indx])
    logger.info(f'The homo-lumo gap is: {hl_gap} eV')

    # Read Files with PDOS info
    xs = [np.loadtxt(files[i]) for i in range(len(files))]
    # Add up all orbitals contribution for each atom type
    ys = np.stack(np.sum(xs[i][:, 3:], axis=1)
                  for i in range(len(files))).transpose()

    if group:
        lig_atoms = 0
        lig_name = ''
        for i in range(len(group)):
            lig_atoms += ys[:, group[i]-1]
            lig_name += legends[group[i]-1]
        ys = np.delete(ys, np.array(group)-1, axis=1)
        ys = np.column_stack((ys, lig_atoms))
        remove_atoms = [legends[i-1] for i in group]
        new_list = list(set(legends) - set(remove_atoms))
        new_list.append(lig_name)
        legends = new_list

    plot_stuff(ys, energies, legends, emin, emax)


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['ligand', 'emin', 'emax']

    return [getattr(args, p) for p in attributes]


# ============<>===============
if __name__ == "__main__":

    msg = "plot_dos.py -emin <Minimum energy window> -emax <Maximum energy window>\
    -ligand <list of atoms to group>"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument('-emin', default=-6, type=float,
                        help='Minimum energy window to plot results')
    parser.add_argument('-emax', default=-1, type=float,
                        help='Maximum energy window to plot results')
    parser.add_argument('-ligand', nargs='+', type=int,
                        help='List of atoms to group')

    main(*read_cmd_line(parser))



"""Compute the nonadiabatic coupling using different methods.

The available methods are:

    * `3-point numerical differentiation <https://doi.org/10.1063/1.4738960>`
    * `levine <dx.doi.org/10.1021/jz5009449>`

Phase correction is also available.

Index
-----
.. currentmodule:: nanoqm.integrals.nonAdiabaticCoupling
.. autosummary::
    calculate_couplings_3points
    calculate_couplings_levine
    compute_overlaps_for_coupling
    correct_phases


API
---
.. autofunction:: calculate_couplings_3points
.. autofunction:: calculate_couplings_levine
.. autofunction:: compute_overlaps_for_coupling
.. autofunction:: correct_phases

"""
__all__ = ['calculate_couplings_3points', 'calculate_couplings_levine',
           'compute_overlaps_for_coupling', 'correct_phases']

import os
import uuid
from os.path import join
from typing import List, Tuple

import numpy as np

from .. import logger
from ..common import (DictConfig, Matrix, MolXYZ, Tensor3D, retrieve_hdf5_data,
                      tuplesXYZ_to_plams)
from ..compute_integrals import compute_integrals_couplings, get_thread_count, get_thread_type


def calculate_couplings_3points(
        dt: float, mtx_sji_t0: Matrix, mtx_sij_t0: Matrix,
        mtx_sji_t1: Matrix, mtx_sij_t1: Matrix) -> Matrix:
    """Calculate the non-adiabatic interaction matrix using 3 geometries.

    see: https://aip.scitation.org/doi/10.1063/1.467455
    the contracted Gaussian functions for the atoms and molecular orbitals coefficients
    are read from a HDF5 File.

    Parameters
    ----------
    dt
        Integration step (atomic units)
    mtx_sji_t0
        Sji Overlap matrix at time t0
    mtx_sij_t0
        SiJ Overlap matrix at time t0
    mtx_sji_t1
        Sji Overlap matrix at time t1
    mtx_sij_t1
        SiJ Overlap matrix at time t1

    Returns
    ------
    np.ndarray
        Coupling matrix

    """
    cte = 1.0 / (4.0 * dt)
    return cte * (3 * (mtx_sji_t1 - mtx_sij_t1) + (mtx_sij_t0 - mtx_sji_t0))


def calculate_couplings_levine(dt: float, w_jk: Matrix,
                               w_kj: Matrix) -> Matrix:
    """Compute coupling using the Levine approximation.

    Compute the non-adiabatic coupling according to:
    `Evaluation of the Time-Derivative Coupling for Accurate Electronic
    State Transition Probabilities from Numerical Simulations`.
    Garrett A. Meek and Benjamin G. Levine.
    dx.doi.org/10.1021/jz5009449 | J. Phys. Chem. Lett. 2014, 5, 2351−2356

    Notes
    -----
        In numpy sinc is defined as sin(pi * x) / (pi * x)


    Parameters
    ----------
    dt
        Integration step (atomic units)
    w_jk
        Overlap matrix
    mtx_sij_t0
        Overlap matrix

    Returns
    -------
    np.ndarray
        Coupling matrix

    """
    # Diagonal matrix
    w_jj = np.diag(np.diag(w_jk))
    w_kk = np.diag(np.diag(w_kj))

    # remove the values from the diagonal
    np.fill_diagonal(w_jk, 0)
    np.fill_diagonal(w_kj, 0)

    # Components A + B
    acos_w_jj = np.arccos(w_jj)
    asin_w_jk = np.arcsin(w_jk)

    a = acos_w_jj - asin_w_jk
    b = acos_w_jj + asin_w_jk
    A = - np.sinc(a / np.pi)
    B = np.sinc(b / np.pi)

    # Components C + D
    acos_w_kk = np.arccos(w_kk)
    asin_w_kj = np.arcsin(w_kj)

    c = acos_w_kk - asin_w_kj
    d = acos_w_kk + asin_w_kj
    C = np.sinc(c / np.pi)
    D = np.sinc(d / np.pi)

    # Components E
    w_lj = np.sqrt(1 - (w_jj ** 2) - (w_kj ** 2))
    w_lk = -(w_jk * w_jj + w_kk * w_kj) / w_lj
    asin_w_lj = np.arcsin(w_lj)
    asin_w_lk = np.arcsin(w_lk)
    asin_w_lj2 = asin_w_lj ** 2
    asin_w_lk2 = asin_w_lk ** 2

    t1 = w_lj * w_lk * asin_w_lj
    x1 = np.sqrt((1 - w_lj ** 2) * (1 - w_lk ** 2)) - 1
    t2 = x1 * asin_w_lk
    t = t1 + t2
    E_nonzero = 2 * asin_w_lj * t / (asin_w_lj2 - asin_w_lk2)

    # Check whether w_lj is different of zero
    E1 = np.where(np.abs(w_lj) > 1e-8, E_nonzero, np.zeros(A.shape))

    # Check whether w_lj is different of zero
    E = np.where(np.isclose(asin_w_lj2, asin_w_lk2), w_lj ** 2, E1)

    cte = 1 / (2 * dt)
    return cte * (np.arccos(w_jj) * (A + B) + np.arcsin(w_kj) * (C + D) + E)


def correct_phases(overlaps: Tensor3D, mtx_phases: Matrix) -> np.ndarray:
    """Correct the phases for all the overlaps."""
    noverlaps = overlaps.shape[0]  # total number of overlap matrices
    dim = overlaps.shape[1]  # Size of the square matrix

    for k in range(noverlaps):
        # Extract phases
        phases_t0, phases_t1 = mtx_phases[k: k + 2]
        phases_t0 = phases_t0.reshape(dim, 1)
        phases_t1 = phases_t1.reshape(1, dim)
        mtx_phases_Sji_t0_t1 = np.dot(phases_t0, phases_t1)

        # Update array with the fixed phases
        overlaps[k] *= mtx_phases_Sji_t0_t1

    return overlaps


def compute_overlaps_for_coupling(
        config: DictConfig,
        pair_molecules: Tuple[MolXYZ, MolXYZ],
        coefficients: Tuple[Matrix, Matrix]) -> Matrix:
    """Compute the Overlap matrices used to compute the couplings.

    Parameters
    ---------
    config
        Configuration of the current task
    pair_molecule
        Molecule to compute the overlap
    coefficients
        Molecular orbital coefficients for each molecule

    Returns
    -------
    Matrix
        containing the overlaps at different times

    """
    # Atomic orbitals overlap
    suv = calcOverlapMtx(config, pair_molecules)

    # Read Orbitals Coefficients
    css0, css1 = coefficients

    return np.dot(css0.T, np.dot(suv, css1))


def read_overlap_data(config: DictConfig, mo_paths: List[str]) -> Tuple[Matrix, Matrix]:
    """Read the Molecular orbital coefficients."""
    mos = retrieve_hdf5_data(config.path_hdf5, mo_paths)

    # Extract a subset of molecular orbitals to compute the coupling
    lowest, highest = compute_range_orbitals(config)
    css0, css1 = tuple(map(lambda xs: xs[:, lowest: highest], mos))

    return css0, css1


def compute_range_orbitals(config: DictConfig) -> Tuple[int, int]:
    """Compute the lowest and highest index used to extract a subset of Columns from the MOs."""
    lowest = config.nHOMO - (config.mo_index_range[0] + config.active_space[0])
    highest = config.nHOMO + config.active_space[1] - config.mo_index_range[0]

    return lowest, highest


def calcOverlapMtx(config: DictConfig, molecules: Tuple[MolXYZ, MolXYZ]) -> Matrix:
    """Compute the overlap matrix between two geometries.

    The calculation of the overlap matrix uses the libint2 library
    at two different geometries: R0 and R1.
    """
    mol_i, mol_j = tuple(tuplesXYZ_to_plams(x) for x in molecules)

    # unique molecular paths
    path_i = join(config["scratch_path"],
                  f"molecule_{uuid.uuid4()}.xyz")
    path_j = join(config["scratch_path"],
                  "molecule_{uuid.uuid4()}.xyz")

    # Write the molecules in atomic units
    mol_i.write(path_i)
    mol_j.write(path_j)

    basis_name = config["cp2k_general_settings"]["basis"]

    thread_count = get_thread_count()
    thread_type = get_thread_type()
    logger.info(f"Will scale over {thread_count} {thread_type} threads")
    try:
        integrals = compute_integrals_couplings(
            path_i, path_j, config["path_hdf5"], basis_name)
    finally:
        os.remove(path_i)
        os.remove(path_j)

    return integrals



"""Nano-QMFlows is a generic python library for computing (numerically) electronic properties \
for nanomaterials like the non-adiabatic coupling vectors (NACV) using several quantum \
chemical (QM) packages."""

from ._version import __version__ as __version__
from ._version_info import version_info as version_info
from ._logger import logger as logger

from .analysis import (
    autocorrelate, dephasing, convolute, func_conv, gauss_function,
    parse_list_of_lists, read_couplings, read_energies,
    read_energies_pyxaid, read_pops_pyxaid, spectral_density
)

from .integrals import (
    calculate_couplings_levine, compute_overlaps_for_coupling)

from .schedule import (calculate_mos, lazy_couplings)

from .workflows import (
    workflow_derivative_couplings, workflow_stddft)

__all__ = [
    'autocorrelate', 'calculate_couplings_levine', 'calculate_mos',
    'compute_overlaps_for_coupling', 'convolute', 'dephasing',
    'func_conv', 'gauss_function', 'lazy_couplings',
    'parse_list_of_lists', 'read_couplings', 'read_energies',
    'read_energies_pyxaid', 'read_pops_pyxaid', 'spectral_density',
    'workflow_derivative_couplings', 'workflow_stddft']



"""Miscellaneous functions to analyze the simulation results.

Index
-----
.. currentmodule:: nanoqm.analysis.tools
.. autosummary::

API
---

"""

import os
from typing import List, Tuple

import numpy as np
import pyparsing as pa
from scipy.optimize import curve_fit

from ..common import fs_to_cm, h2ev, hbar, r2meV

""" Functions to fit data """


def gauss_function(x: float, sigma: float) -> np.ndarray:
    """Compute a Gaussian function used for fitting data."""
    return np.exp(-0.5 * (-x / sigma) ** 2)


def lorentzian_function(x_L, sigma_L, amplitude_L):
    """Compute a Lorentzian function used for fitting data."""
    return amplitude_L * sigma_L**2 / (sigma_L**2 + x_L ** 2)


def exp_function(x, x0, amplitude):
    """Compute an Exponential function used for fitting data."""
    return amplitude * np.exp(-x / x0)


def sine_function(t_phonon, amplitude, offset, phase, n_periods, dt):
    """Compute a sinusoidal function used for fitting data."""
    t = np.arange(0, n_periods * t_phonon, dt)
    # (gap) energy
    y = offset + amplitude * (np.sin(2 * np.pi * t / t_phonon + phase))
    y_mean = np.mean(y)
    y_dummy = y / h2ev    # to delete ...
    return y_dummy, y, y_mean, t


def sqrt_func(x, a):
    """Compute a square root function used for fitting data."""
    return a * np.sqrt(x)


def func_conv(x_real: np.ndarray, x_grid: np.ndarray, delta: float) -> np.ndarray:
    """Compute a convolution on a grid using a Gaussian function."""
    return np.exp(-2 * (x_grid - x_real) ** 2 / delta ** 2)


def convolute(x: np.ndarray, y: np.ndarray, x_points: np.ndarray, sigma: float) -> np.ndarray:
    """Convolute a spectrum on a grid of x_points.

    You need as input x, y and the grid where to convolute.
    """
    # Compute gaussian prefactor
    prefactor = np.sqrt(2.0) / (sigma * np.sqrt(np.pi))
    # Convolute spectrum over grid
    y_points = prefactor * np.stack([
        np.sum(y * func_conv(x, x_point, sigma)) for x_point in x_points
    ])
    return y_points


""" Useful functions to compute autocorrelation, dephasing, etc. """


def autocorrelate(f: np.ndarray) -> Tuple[float, float]:
    """Compute the un-normalized and normalized autocorrelation of a function."""
    d_f = f - f.mean()
    d_f2 = np.append(d_f, d_f, axis=0)
    # Compute the autocorrelation function
    uacf = np.correlate(d_f, d_f2, "valid")[:d_f.size] / d_f.size
    # Compute the normalized autocorrelation function
    nacf = uacf / uacf[0]
    return uacf, nacf


def spectral_density(f, dt):
    """Fourier Transform of a given function f using a dense grid with 100000 points.

    In the case of a FFT of a normalized autocorrelation function,
    this corresponds to a spectral density
    """
    n_pts = 100000
    f_fft = abs(1 / np.sqrt(2 * np.pi) * np.fft.rfft(f, n_pts) * dt) ** 2
    # Fourier Transform of the time axis
    freq = np.fft.rfftfreq(n_pts, dt)
    # Conversion of the x axis (given in cycles/fs) to cm-1
    freq = freq * fs_to_cm
    return f_fft, freq


def dephasing(f: np.ndarray, dt: float):
    """Compute the dephasing time of a given function.

        f            = energies, in eV  (if other function/unit: then the line_broadening will not be in eV)
        dt           = time step, in fs

    Use the optical response formalisms:
    S. Mukamel, Principles of Nonlinear Optical Spectroscopy, 1995
    About the implementation we use the 2nd order cumulant expansion.
    See also eq. (2) in : Kilina et al. Phys. Rev. Lett., 110, 180404, (2013)
    To calculate the dephasing time tau we fit the dephasing function to a
    gaussian of the type : exp(-0.5 * (-x / tau) ** 2)
    """
    # Conversion of hbar to hartree * fs
    hbar_au = hbar / h2ev
    ts = np.arange(f.shape[0]) * dt
    cumu_ii = np.asarray([np.trapz(f[0:i + 1], dx=(dt / hbar), axis=0) for i in range(ts.size)])
    cumu_i = np.asarray([np.trapz(cumu_ii[0:i + 1], dx=(dt / hbar), axis=0)
                         for i in range(ts.size)])
    deph = np.exp(-cumu_i)

    return deph, ts


def fit_dephasing(fit_func, deph, ts, res, t_deph_guess):
    """Work in progress (?).

    fit_func = 0 for Gaussian fit, or 1 for exponential fit
    deph     = dephasing function vs. time "
    ts       = time, in fs
    res      = factor by which to increase the time resolution of ts for the fit
    t_deph_guess = initial guess, in fs
    """
    np.seterr(over='ignore')

    if fit_func == 0:
        # fit with Gaussian
        popt, pcov = curve_fit(gauss_function, ts, deph, p0=(t_deph_guess, deph[0]))    # [0]
        ts_fit = np.arange(res * deph.shape[0]) * dt / res
        deph_fit = popt[1] * np.exp(-0.5 * (-ts_fit / popt[0]) ** 2)
        deph_time = popt[0]                                           # in fs (defined as standard deviation of a Gaussian)
        e_fwhm = std_to_fwhm * hbar / deph_time                       # FWHM in eV
        perr = np.sqrt(np.diag(pcov))
        deph_time_err = perr[0]                                            # error (standard deviation) for the deph. time
        e_fwhm_err = deph_time_err * std_to_fwhm * hbar / (deph_time ** 2)  # error (standard deviation) for the FWHM
    elif fit_func == 1:
        # fit with exponential
        popt, pcov = curve_fit(exp_function, ts, deph, p0=(t_deph_guess, deph[0]))  # [0]
        ts_fit = np.arange(res * deph.shape[0]) * dt / res
        deph_fit = popt[1] * np.exp(-ts_fit / popt[0])
        deph_time = popt[0]                                           # in fs (defined as the exp. time constant)
        e_fwhm = 2 * hbar / deph_time                                 # FWHM in eV
        perr = np.sqrt(np.diag(pcov))
        deph_time_err = perr[0]                                       # error (standard deviation) for the deph. time
        e_fwhm_err = deph_time_err * 2 * hbar / (deph_time ** 2)      # error (standard deviation) for the FWHM

    return ts_fit, deph_fit, deph_time, deph_time_err, e_fwhm, e_fwhm_err


def read_couplings(path_hams, ts):
    """Read the non adiabatic coupling vectors from the files generated for the NAMD simulations."""
    files_im = [os.path.join(path_hams, f'Ham_{i}_im')
                for i in range(ts)]
    xs = np.stack(np.loadtxt(fn) for fn in files_im)
    return xs * r2meV  # return energies in meV


def read_energies(path_hams, ts):
    """Read the molecular orbital energies of each state.

    The target files are generated for the NAMD simulations.
    """
    files_re = [os.path.join(path_hams, f'Ham_{i}_re')
                for i in range(ts)]
    xs = np.stack(np.diag(np.loadtxt(fn)) for fn in files_re)
    return xs * r2meV / 1000  # return energies in eV


def read_energies_pyxaid(path, fn, nstates, nconds):
    """Read the molecular orbital energies of each state from the output files generated by PYXAID."""
    inpfile = os.path.join(path, fn)
    cols = tuple(range(5, nstates * 2 + 5, 2))
    xs = np.stack(np.loadtxt(f'{inpfile}{j}', usecols=cols)
                  for j in range(nconds)).transpose()
    # Rows = timeframes ; Columns = states ; tensor = initial conditions
    xs = xs.swapaxes(0, 1)
    return xs


def read_pops_pyxaid(path, fn, nstates, nconds):
    """Read the population of each state from the output files generated by PYXAID."""
    inpfile = os.path.join(path, fn)
    cols = tuple(range(3, nstates * 2 + 3, 2))
    xs = np.stack(np.loadtxt(f'{inpfile}{j}', usecols=cols)
                  for j in range(nconds)).transpose()
    # Rows = timeframes ; Columns = states ; tensor = initial conditions
    xs = xs.swapaxes(0, 1)
    return xs


def parse_list_of_lists(xs: str) -> List[List[int]]:
    """Parse a list of list of integers using pyparsing."""
    enclosed = pa.Forward()  # Parser to be defined later
    natural = pa.Word(pa.nums)  # Natural Number
    # Nested Grammar
    nestedBrackets = pa.nestedExpr(pa.Suppress(
        '['), pa.Suppress(']'), content=enclosed)
    enclosed << (natural | pa.Suppress(',') | nestedBrackets)
    try:
        rs = enclosed.parseString(xs).asList()[0]
        return list(map(lambda x: list(map(int, x)), rs))
    except pa.ParseException:
        raise RuntimeError("Invalid Macro states Specification")



"""Schedule API."""
from .scheduleCoupling import (compute_phases, lazy_couplings,
                               write_hamiltonians)
from .scheduleCP2K import (prepare_cp2k_settings, prepare_job_cp2k)
from .components import (calculate_mos, create_point_folder, split_file_geometries)

__all__ = ['calculate_mos', 'compute_phases',
           'create_point_folder', 'lazy_couplings',
           'prepare_cp2k_settings', 'prepare_job_cp2k',
           'split_file_geometries', 'write_hamiltonians']



"""Module to configure and run CP2K jobs.

Index
-----
.. currentmodule:: nanoqm.schedule.scheduleCP2K
.. autosummary::
    prepare_job_cp2k

API
---
.. autofunction:: prepare_job_cp2k

"""

from __future__ import annotations

import fnmatch
import os
from os.path import join
from pathlib import Path
from typing import Any, Dict

from noodles import schedule  # Workflow Engine
from qmflows import Settings, cp2k, templates
from qmflows.packages import CP2K, CP2K_Result
from qmflows.parsers import string_to_plams_Molecule
from qmflows.type_hints import PromisedObject

from .. import logger


def try_to_read_wf(path_dir: str | os.PathLike[str]) -> str:
    """Try to get a wave function file from ``path_dir``.

    Returns
    -------
    str
        Path to the wave function file.

    Raises
    ------
    RuntimeError
        If there is not a wave function file.

    """
    xs = os.listdir(path_dir)
    files = list(filter(lambda x: fnmatch.fnmatch(x, '*wfn'), xs))
    if files:
        return join(path_dir, files[0])
    else:
        msg = f"There are no wave function file in path: {os.fspath(path_dir)!r}\n"
        msg += print_cp2k_error(path_dir, "err")
        msg += print_cp2k_error(path_dir, "out")
        raise RuntimeError(msg)


def prepare_cp2k_settings(
        settings: Settings, dict_input: Dict[str, Any], guess_job: CP2K_Result) -> Settings:
    """Fill in the parameters for running a single job in CP2K.

    Parameters
    ----------
    settings
        Input for CP2K
    dict_input
        Input for the current molecular geometry
    guess_job
        Previous job to read the guess wave function

    Returns
    .......
    CP2K
        job to run

    """
    dft = settings.specific.cp2k.force_eval.dft
    dft['print']['mo']['filename'] = dict_input["job_files"].get_MO

    # Global parameters for CP2K
    settings.specific.cp2k['global']['project'] = f'point_{dict_input["k"]}'

    if guess_job is not None:
        plams_dir = guess_job.archive['plams_dir']
        if plams_dir is None:
            raise RuntimeError(f"There are no wave function file in path: None\n")
        dft.wfn_restart_file_name = try_to_read_wf(plams_dir)

    input_args = templates.singlepoint.overlay(settings)

    return input_args


@schedule
def prepare_job_cp2k(
        settings: Settings, dict_input: Dict[str, Any], guess_job: PromisedObject) -> CP2K:
    """Generate a :class:`qmflows.packages.CP2K` job.

    Parameters
    ----------
    settings
        Input for CP2K
    dict_input
        Input for the current molecular geometry
    guess_job
        Previous job to read the guess wave function

    Returns
    -------
    :class:`qmflows.packages.CP2K`
        job to run

    """
    job_settings = prepare_cp2k_settings(settings, dict_input, guess_job)

    # remove keywords not use on the next translation phase
    for x in ('basis', 'potential'):
        if x in job_settings:
            del job_settings[x]

    return cp2k(
        job_settings, string_to_plams_Molecule(dict_input["geometry"]),
        work_dir=dict_input['point_dir'])


def print_cp2k_error(path_dir: str | os.PathLike[str], prefix: str) -> str:
    """Search for error in the CP2K output files."""
    err_file = next(Path(path_dir).glob(f"*{prefix}"), None)
    msg = ""
    if err_file is not None:
        with open(err_file, 'r') as handler:
            err = handler.read()
        msg = f"CP2K {prefix} file:\n{err}\n"
        logger.error(msg)

    return msg



#!/usr/bin/env python
"""This program plots several properties related to the interaction of a pair of states s1 and s2.

1. The energies of the two states along the MD trajectory
2. The non-adiabatic coupling between the two states (in meV)
3. The normalized autocorrelation function (NAUF) of a pair of entangled states during
the MD trajectory.
   It describes how the interaction between two states depend on earlier time.
   The AUF of poorly correlated states goes rapidly to zero.
4. The unnormalized autocorrelation function. It provides similar infos than NAUF,
 but the starting value is average oscillation at t=0 (in eV**2)
5. The spectral density computed from the fourier transform of the NAUF.
 It identifies which phonones coupled to the electronic subsystem.
 In this case, it tells which phonons are involved in the coupling between
 the pair of entangles states.
6. The dephasing function. It indicates how the two entangled states decohere over time.
   The inverse of the decoherence rate is related to the homogenoues line broadening of
   emission lines when computed between HOMO and LUMO states.

Note that you have to provide the location of the folder where the NAMD hamiltonian
 elements are stored using the -p flag.
"""

import argparse
import glob
import os

import matplotlib.pyplot as plt
import numpy as np

from nanoqm import logger
from nanoqm.analysis import (autocorrelate, dephasing, read_couplings,
                             read_energies, spectral_density)
from nanoqm.common import fs_to_nm


def plot_stuff(ens, coupls, acf, sd, deph, rate, s1, s2, ts, wsd, wdeph, dt):
    """
    arr - a vector of y-values that are plot
    plot_mean, save_plot - bools telling to plot the mean and save the plot or not,
    respectively
    """
    dim_x = np.arange(ts) * dt

    ax1 = plt.subplot(321)
    ax1.set_xlabel('Time (fs)')
    ax1.set_ylabel('Energy (eV)')
    ax1.plot(dim_x, ens[:, 0], c='r')
    ax1.plot(dim_x, ens[:, 1], c='b')

    ax2 = plt.subplot(323)
    ax2.set_xlabel('Time (fs)')
    ax2.set_ylabel('Normalized AUF')
    ax2.set_ylim(-1, 1)
#    ax2.plot(ts, acf[:, 1, 0], c='r')
#    ax2.plot(ts, acf[:, 1, 1], c='b')
    ax2.plot(dim_x, acf[:, 1, 2], c='g')
    ax2.axhline(0, c="black")

    ax3 = plt.subplot(324)
    ax3.set_xlabel('Time (fs)')
    ax3.set_ylabel('Un-normalized AUF')
#    ax3.plot(ts, acf[:, 1, 0], c='r')
#    ax3.plot(ts, acf[:, 1, 1], c='b')
    ax3.plot(dim_x, acf[:, 0, 2], c='g')
    ax3.axhline(0, c="black")

    ax4 = plt.subplot(326)
    ax4.set_xlabel('Time (fs)')
    ax4.set_ylabel('Dephasing (arbitrary units)')
    ax4.set_xlim(0, wdeph)
    ax4.plot(dim_x, deph[:, 0], c='r')
    ax4.plot(dim_x, deph[:, 1], c='b')

    ax5 = plt.subplot(325)
    ax5.set_xlabel('Frequency (cm-1)')
    ax5.set_ylabel('Spectral Density (arbitrary units)')
    ax5.set_xlim(0, wsd)
#    ax5.plot(sd[:, 1, 0], sd[:, 0, 0], c='r')
#    ax5.plot(sd[:, 1, 1], sd[:, 0, 1], c='b')
    ax5.plot(sd[:, 1, 2], sd[:, 0, 2], c='g')
    logger.info(f'The dephasing time is : {rate:f} fs')
    logger.info(f'The homogenous line broadening is  : {1 / rate * fs_to_nm:f} nm')
    # Conversion 1 fs = 4.13567 eV
    lb = 1 / rate * 4.13567
    logger.info(f'The homogenous line broadening is  : {lb:f} eV')

    ax6 = plt.subplot(322)
    ax6.set_xlabel('Time (fs)')
    ax6.set_ylabel('Coupling (meV)')
    ax6.plot(dim_x, coupls[:, s1, s2], c='b')
    av_coupl = np.average(abs(coupls[:, s1, s2]))
    ax6.axhline(av_coupl, c="black")
    logger.info(f'The average coupling strength is : {av_coupl:f} meV')

    fileName = "MOs.png"
    plt.savefig(fileName, format='png', dpi=300)

    plt.show()


def main(path_hams, s1, s2, dt, ts, wsd, wdeph):
    if ts == 'All':
        files = glob.glob(os.path.join(path_hams, 'Ham_*_re'))
        ts = len(files)
    else:
        ts = int(ts)
    energies = read_energies(path_hams, ts)
    couplings = read_couplings(path_hams, ts)
    # Compute the energy difference between pair of states
    d_E = energies[:, s1] - energies[:, s2]
    # Generate a matrix with s1, s2 and diff between them
    en_states = np.column_stack((energies[:, s1], energies[:, s2], d_E))
    # Compute autocorrelation function for each column (i.e. state)
    acf = np.stack(
        autocorrelate(en_states[:, i]) for i in range(en_states.shape[1])).transpose()
    # Compute the spectral density for each column using the normalized acf
    sd = np.stack(
        spectral_density(acf[:, 1, i], dt) for i in range(en_states.shape[1])).transpose()
    # Compute the dephasing time for the uncorrelated acf between two states
    deph, rate = dephasing(acf[:, 0, 2], dt)
    # Plot stuff
    plot_stuff(
        en_states, couplings, acf, sd, deph, rate, s1, s2, ts, wsd, wdeph, dt)


def read_cmd_line(parser):
    """
    Parse Command line options.
    """
    args = parser.parse_args()

    attributes = ['p', 's1', 's2', 'dt', 'ts', 'wsd', 'wdeph']

    return [getattr(args, p) for p in attributes]


if __name__ == "__main__":
    msg = "plot_decho -p <path/to/hamiltonians> -s1 <State 1> -s2 <State 2>\
     -dt <timestep>\
     -ts <time window for analysis>\
     -wsd <energy window for spectral density plot in cm-1>\
      -wdeph <time window for dephasing time in fs>"

    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument(
        '-p', required=True, help='path to the Hamiltonian files in Pyxaid format')
    parser.add_argument(
        '-s1', required=True, type=int, help='Index of the first state')
    parser.add_argument(
        '-s2', required=True, type=int, help='Index of the second state')
    parser.add_argument(
        '-dt', required=True, type=float, help='Timestep for MD simulation')
    parser.add_argument(
        '-ts', type=str, default='All', help='Time range for plot')
    parser.add_argument(
        '-wsd', type=int, default=1500,
        help='energy window for spectral density plot in cm-1')
    parser.add_argument(
        '-wdeph', type=int, default=50, help='time window for dephasing time in fs')

    main(*read_cmd_line(parser))



#! /usr/bin/env python

"""
This program merges the HDF5 files obtained from the SCF calculations when
the MD trajectory has been split in more than one block.
Example:

mergeHDF5.py -i chunk_a.hdf5 chunk_b.hdf5 chunk_c.hdf5 -o total.hdf5

An empty total.hdf5 file should be already available before using the script.
"""
import argparse
import h5py
import os

# ====================================<>=======================================
msg = " script -i <Path(s)/to/source/hdf5> -o <path/to/destiny/hdf5>"

parser = argparse.ArgumentParser(description=msg)
parser.add_argument('-i', required=True,
                    help='Path(s) to the HDF5 to merge', nargs='+')
parser.add_argument('-o', required=True,
                    help='Path to the HDF5 were the merge is going to be stored')


def read_cmd_line():
    """
    Parse Command line options.
    """
    args = parser.parse_args()
    inp = args.i
    out = args.o

    return inp, out
# ===============><==================


def mergeHDF5(inp, out):
    """Merge Recursively two hdf5 Files."""
    with h5py.File(inp, 'r') as f5, h5py.File(out, 'r+') as g5:
        merge_recursively(f5, g5)


def merge_recursively(f, g):
    """Traverse all the groups tree and copy the different datasets."""
    for k in f.keys():
        if k not in g:
            if isinstance(f[k], h5py.Dataset):
                f.copy(k, g)
            else:
                g.create_group(k)
                merge_recursively(f[k], g[k])
        elif isinstance(f[k], h5py.Group):
            merge_recursively(f[k], g[k])


def main():
    inps, out = read_cmd_line()
    if not os.path.exists(out):
        touch(out)
    for i in inps:
        mergeHDF5(i, out)


def touch(fname, times=None):
    """Equivalent to unix touch command"""
    with open(fname, 'a'):
        os.utime(fname, times)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Comman line interface to run the workflows.

Usage:
    run_workflow.py -i input.yml

Available workflow:
    * absorption_spectrum
    * derivative_couplings
    * single_points
    * ipr_calculation
    * coop_calculation

"""

import argparse
import os

import yaml

from .. import logger
from ..common import UniqueSafeLoader
from .input_validation import process_input
from .workflow_coop import workflow_crystal_orbital_overlap_population
from .workflow_coupling import workflow_derivative_couplings
from .workflow_ipr import workflow_ipr
from .workflow_single_points import workflow_single_points
from .workflow_stddft_spectrum import workflow_stddft

msg = "run_workflow.py -i input.yml"

parser = argparse.ArgumentParser(description=msg)
parser.add_argument('-i', required=True,
                    help="Input file in YAML format")


dict_workflows = {
    'absorption_spectrum': workflow_stddft,
    'derivative_couplings': workflow_derivative_couplings,
    'single_points': workflow_single_points,
    'ipr_calculation': workflow_ipr,
    'coop_calculation': workflow_crystal_orbital_overlap_population}


def main():
    """Parse the command line arguments and run workflow."""
    args = parser.parse_args()
    input_file = args.i
    with open(input_file, 'r') as f:
        dict_input = yaml.load(f, Loader=UniqueSafeLoader)
    if 'workflow' not in dict_input:
        raise RuntimeError(
            "The name of the workflow is required in the input file")
    else:
        workflow_name = dict_input['workflow']

    # Read and process input
    inp = process_input(input_file, workflow_name)

    # run workflow
    function = dict_workflows[workflow_name]

    logger.info(f"Running worflow using: {os.path.abspath(input_file)}")
    function(inp)


if __name__ == "__main__":
    main()



"""Functionality to check that the input provided by the user is valid.

Index
-----
.. currentmodule:: nanoqm.workflows.input_validation
.. autosummary:: process_input

API
---
.. autofunction:: process_input

"""

import os
import warnings
from os.path import join
from pathlib import Path
from typing import Any, Dict, Union

import yaml
from schema import SchemaError
from scm.plams import Molecule

from qmflows import Settings
from qmflows.type_hints import PathLike

from .. import logger
from ..common import DictConfig, UniqueSafeLoader, valence_electrons
from .schemas import (schema_absorption_spectrum, schema_coop,
                      schema_cp2k_general_settings,
                      schema_derivative_couplings,
                      schema_distribute_absorption_spectrum,
                      schema_distribute_derivative_couplings,
                      schema_distribute_single_points, schema_ipr,
                      schema_single_points)
from .templates import create_settings_from_template

schema_workflows = {
    'absorption_spectrum': schema_absorption_spectrum,
    'derivative_couplings': schema_derivative_couplings,
    'single_points': schema_single_points,
    'cp2k_general_settings': schema_cp2k_general_settings,
    'distribute_derivative_couplings': schema_distribute_derivative_couplings,
    'distribute_absorption_spectrum': schema_distribute_absorption_spectrum,
    'distribute_single_points': schema_distribute_single_points,
    'ipr_calculation': schema_ipr,
    'coop_calculation': schema_coop
}


def process_input(input_file: PathLike, workflow_name: str) -> DictConfig:
    """Read the `input_file` in YAML format and validate it.

    Use the corresponding `workflow_name` schema and return a nested
    dictionary with the input.

    Parameters
    ----------
    input_file
        path to the input

    Returns
    -------
    dict
        Configuration to run the given workflow
    Raises
    ------
    SchemaError
        If the input is not valid

    """
    schema = schema_workflows[workflow_name]

    with open(input_file, 'r') as f:
        dict_input = yaml.load(f.read(), Loader=UniqueSafeLoader)

    try:
        d = schema.validate(dict_input)
        return DictConfig(InputSanitizer(d).sanitize())

    except SchemaError as e:
        msg = f"There was an error in the input yaml provided:\n{e}"
        logger.warning(msg)
        raise


class InputSanitizer:
    """Class to sanitize the input."""

    def __init__(self, input_dict: Dict):
        """Set the class properties."""
        self.user_input = input_dict
        self.general = input_dict["cp2k_general_settings"]

    def sanitize(self) -> Dict:
        """Apply all the sanity check on the input."""
        self.create_settings()
        self.apply_templates()
        self.add_missing_keywords()
        self.add_executable()
        self.print_final_input()

        return self.user_input

    def create_settings(self) -> None:
        """Transform the CP2K input dict into :class:`QMFLows.Settings`."""
        self.general['cp2k_settings_main'] = Settings(self.general['cp2k_settings_main'])
        self.general['cp2k_settings_guess'] = Settings(self.general['cp2k_settings_guess'])

    def apply_templates(self) -> None:
        """Apply a template for CP2K if the user requested it."""
        for s in [
                self.general[x] for x in ('cp2k_settings_main', 'cp2k_settings_guess')]:
            val = s['specific']

            if "template" in val:
                cp2k_template = create_settings_from_template(
                    self.general, val['template'], self.user_input["path_traj_xyz"])
                # remove template
                del s['specific']['template']

                # Add other keywords
                s['specific'] = cp2k_template.overlay(s['specific'])

    def add_executable(self) -> None:
        """Add executable to the job settings."""
        self.general['cp2k_settings_main']['executable'] = self.general['executable']
        self.general['cp2k_settings_guess']['executable'] = self.general['executable']

    def add_missing_keywords(self) -> None:
        """Add missing input data using the defaults."""
        # Add the `added_mos` and `mo_index_range` keywords
        if self.user_input.get('nHOMO') is None:
            self.user_input["nHOMO"] = self.compute_homo_index()

        # Added_mos keyword
        if self.user_input.get('compute_orbitals'):
            self.add_mo_index_range()
        else:
            logger.info("Orbitals are neither print nor store!")

        # Add restart point provided by the user
        self.add_restart_point()

        # Add basis sets
        self.add_basis()

        # add cell parameters
        self.add_cell_parameters()

        # Add Periodic properties
        self.add_periodic()

        # Add charge
        self.add_charge()

        # Add multiplicity
        self.add_multiplicity()

        # Add DFT exchange part
        self.add_functional_x()

        # Add DFT correlation part
        self.add_functional_c()

    def compute_homo_index(self) -> int:
        """Compute the HOMO index."""
        charge = self.general['charge']
        mol = Molecule(self.user_input["path_traj_xyz"], 'xyz')

        number_of_electrons = sum(valence_electrons[at.symbol] for at in mol.atoms)
        # Correct for total charge of the system
        number_of_electrons = number_of_electrons - charge

        if (number_of_electrons % 2) != 0:
            warnings.warn(
                "Unpair number of electrons detected when computing the HOMO")

        return (number_of_electrons // 2) + (number_of_electrons % 2)

    def add_basis(self) -> None:
        """Add path to the basis and potential."""
        setts = [self.general[p] for p in ['cp2k_settings_main', 'cp2k_settings_guess']]

        root = os.path.abspath(self.general['path_basis'])
        if self.general['potential_file_name'] is not None:
            names = [join(root, f) for f in self.general["potential_file_name"]]
        else:
            names = [join(root, "GTH_POTENTIALS")]

        # add basis and potential path
        if self.general["path_basis"] is not None:
            logger.info("path to basis added to cp2k settings")
            for x in setts:
                x.basis = self.general['basis']
                x.potential = self.general['potential']

                # Do not overwrite explicitly specified CP2K settings
                dft = x.specific.cp2k.force_eval.dft
                if dft.get("potential_file_name") is None:
                    dft["potential_file_name"] = names

                # Choose the file basis to use
                self.select_basis_file(x)

    def select_basis_file(self, sett: Settings) -> None:
        """Choose the right basis set based on the potential and basis name."""
        dft = sett.specific.cp2k.force_eval.dft

        # Do not overwrite explicitly specified CP2K settings
        if dft.get("basis_set_file_name") is not None:
            return

        root = os.path.abspath(self.general['path_basis'])
        if self.general['basis_file_name'] is not None:
            dft["basis_set_file_name"] = [join(root, f) for f in self.general["basis_file_name"]]
        else:
            dft["basis_set_file_name"] = [join(root, "BASIS_MOLOPT")]
            # USE ADMM
            if dft.xc.get("xc_functional pbe") is None:
                dft["basis_set_file_name"] += [
                    join(root, "BASIS_ADMM_MOLOPT"),
                    join(root, "BASIS_ADMM"),
                ]

    def add_cell_parameters(self) -> None:
        """Add the Unit cell information to both the main and the guess settings."""
        # Search for a file containing the cell parameters
        for s in (self.general[p] for p in [
                'cp2k_settings_main',
                'cp2k_settings_guess']):
            if self.general["file_cell_parameters"] is None:
                s.cell_parameters = self.general['cell_parameters']
                s.cell_angles = None
            else:
                s.cell_parameters = None
            s.cell_angles = None

    def add_periodic(self) -> None:
        """Add the keyword for the periodicity of the system."""
        for s in (
            self.general[p] for p in [
                'cp2k_settings_main',
                'cp2k_settings_guess']):
            s.specific.cp2k.force_eval.subsys.cell.periodic = self.general['periodic']

    def add_charge(self) -> None:
        """Add the keyword for the charge of the system."""
        for s in (
            self.general[p] for p in [
                'cp2k_settings_main',
                'cp2k_settings_guess']):
            s.specific.cp2k.force_eval.dft.charge = self.general['charge']

    def add_multiplicity(self) -> None:
        """Add the keyword for the multiplicity of the system only if greater than 1."""
        if self.general['multiplicity'] > 1:
            for s in (
                    self.general[p] for p in ('cp2k_settings_main', 'cp2k_settings_guess')):
                s.specific.cp2k.force_eval.dft.multiplicity = self.general['multiplicity']
                s.specific.cp2k.force_eval.dft.uks = ""

    def add_functional_x(self) -> None:
        """Add the keyword for the exchange part of the DFT functional: GGA or MGGA."""
        if self.general['functional_x'] is None:
            return
        for s in (
            self.general[p] for p in [
                'cp2k_settings_main',
                'cp2k_settings_guess']):
            s.specific.cp2k.force_eval.dft.xc.xc_functional[self.general['functional_x']] = {}

    def add_functional_c(self) -> None:
        """Add the keyword for the correlation part of the DFT functional: GGA or MGGA."""
        if self.general['functional_c'] is None:
            return
        for s in (
            self.general[p] for p in [
                'cp2k_settings_main',
                'cp2k_settings_guess']):
            s.specific.cp2k.force_eval.dft.xc.xc_functional[self.general['functional_c']] = {}

    def add_restart_point(self) -> None:
        """Add a restart file if the user provided it."""
        guess = self.general['cp2k_settings_guess']
        wfn = self.general['wfn_restart_file_name']
        if wfn is not None and wfn:
            dft = guess.specific.cp2k.force_eval.dft
            dft.wfn_restart_file_name = Path(wfn).absolute().as_posix()

    def add_mo_index_range(self) -> None:
        """Compute the MO range to print."""
        active_space = self.user_input["active_space"]
        nHOMO = self.user_input["nHOMO"]
        mo_index_range = nHOMO - active_space[0], nHOMO + active_space[1]
        self.user_input["mo_index_range"] = mo_index_range

        # mo_index_range keyword
        cp2k_main = self.general['cp2k_settings_main']
        dft_main_print = cp2k_main.specific.cp2k.force_eval.dft.print
        dft_main_print.mo.mo_index_range = "{} {}".format(
            mo_index_range[0] + 1, mo_index_range[1])

        # added_mos
        cp2k_main.specific.cp2k.force_eval.dft.scf.added_mos = mo_index_range[1] - nHOMO

        # Add section to Print the orbitals
        mo = cp2k_main.specific.cp2k.force_eval.dft.print.mo
        mo.add_last = "numeric"
        mo.each.qs_scf = 0
        mo.eigenvalues = ""
        mo.eigenvectors = ""
        mo.ndigits = 36

    def print_final_input(self) -> None:
        """Print the input after post-processing."""
        xs = self.user_input.copy()

        for k, v in self.user_input.items():
            xs[k] = recursive_traverse(v)

        with open("input_parameters.yml", "w") as f:
            yaml.dump(xs, f, indent=4)


def recursive_traverse(val: Union[Dict, Settings, Any]) -> Union[Dict, Settings, Any]:
    """Check if the value of a key is a Settings instance a transform it to plain dict."""
    if isinstance(val, dict):
        if isinstance(val, Settings):
            return val.as_dict()
        else:
            return {k: recursive_traverse(v) for k, v in val.items()}
    else:
        return val



"""Initial configuration setup.

Index
-----
.. currentmodule:: nanoqm.workflows.initialization
.. autosummary::
    initialize
    split_trajectory
API
---
.. autofunction:: initialize
.. autofunction:: split_trajectory

"""

from __future__ import annotations

__all__ = ['initialize', 'read_swaps', 'split_trajectory']

import re
import fnmatch
import getpass
import os
import subprocess
import tempfile
from os.path import join
from pathlib import Path
from subprocess import PIPE, Popen
from typing import List, Union

import h5py
import numpy as np
import qmflows
from nanoutils import RecursiveItemsView
from packaging.version import Version
from qmflows.common import AtomBasisKey
from qmflows.parsers import parse_string_xyz
from qmflows.parsers.cp2k import read_cp2k_basis
from qmflows.type_hints import PathLike

from .. import logger, __version__, __path__ as nanoqm_path
from .._logger import EnableFileHandler
from ..common import (BasisFormats, DictConfig, Matrix, change_mol_units,
                      is_data_in_hdf5, retrieve_hdf5_data,
                      store_arrays_in_hdf5)
from ..schedule.components import create_point_folder, split_file_geometries


def initialize(config: DictConfig) -> DictConfig:
    """Initialize all the data required to schedule the workflows.

    Returns
    -------
    DictConfig
        Input to run the workflow.

    """
    with EnableFileHandler(f'{config.project_name}.log'):
        logger.info(f"Using nano-qmflows version: {qmflows.__version__} ")
        logger.info(f"nano-qmflows path is: {nanoqm_path[0]}")
        logger.info(f"Working directory is: {os.path.abspath('.')}")
        logger.info(f"Data will be stored in HDF5 file: {config.path_hdf5}")

        # Scratch folder
        scratch_path = create_path_option(config["scratch_path"])
        if scratch_path is None:
            scratch_path = (
                Path(tempfile.gettempdir()) / getpass.getuser() / config.project_name
            )
            logger.warning(f"path to scratch was not defined, using: {scratch_path}")
        config['workdir'] = scratch_path

        # If the directory does not exist create it
        if not scratch_path.exists():
            scratch_path.mkdir(parents=True)

        # Touch HDF5 if it doesn't exists
        if not os.path.exists(config.path_hdf5):
            Path(config.path_hdf5).touch()

        # all_geometries type :: [String]
        geometries = split_file_geometries(config["path_traj_xyz"])
        config['geometries'] = geometries

        # Create a folder for each point the the dynamics
        enumerate_from = config["enumerate_from"]
        len_geometries = len(geometries)
        config["folders"] = create_point_folder(
            scratch_path, len_geometries, enumerate_from)

        config['calc_new_wf_guess_on_points'] = guesses_to_compute(
            config['calculate_guesses'], enumerate_from, len_geometries)

        # Generate a list of tuples containing the atomic label
        # and the coordinates to generate the primitive CGFs
        atoms = parse_string_xyz(geometries[0])
        if 'angstrom' in config["geometry_units"].lower():
            atoms = change_mol_units(atoms)

        # Save Basis to HDF5
        save_basis_to_hdf5(config)

    return config


def save_basis_to_hdf5(config: DictConfig) -> None:
    """Store the specification of the basis set in the HDF5 to compute the integrals."""
    root: str = config['cp2k_general_settings']['path_basis']
    files: "None | list[str]" = config['cp2k_general_settings']['basis_file_name']
    if files is not None:
        basis_paths = [os.path.join(root, i) for i in files]
    else:
        basis_paths = [os.path.join(root, "BASIS_MOLOPT")]

    for path in basis_paths:
        if not is_data_in_hdf5(config.path_hdf5, path):
            store_cp2k_basis(config.path_hdf5, path)


_NANO_QMFLOWS_0_13_0 = Version("0.13.0")

_BASIS_PATH_PATTERN = re.compile(r"/cp2k/basis/([a-z]+)/([^/]+)/(coefficients|exponents)")
_SUFFIX_PATTERN = re.compile(r"/(coefficients|exponents)")


def _join(xs: AtomBasisKey, *args: str) -> str:
    return join("cp2k/basis", xs.atom.lower(), xs.basis, *args)


def _prepend_exp_set(match: re.Match) -> str:
    return f"/0/{match[1]}"


def _convert_legacy_basis(f: h5py.Group) -> None:
    """Update the basis-set format to nanoqm 0.13.0-style and store ``nanoqm__version__``."""
    items = [(k, v) for k, v in RecursiveItemsView(f) if _BASIS_PATH_PATTERN.match(k) is not None]
    for name, dset in items:
        # Update the dataset-path
        name_new = _SUFFIX_PATTERN.sub(_prepend_exp_set, name)
        f.move(name, name_new)

        # Update the type of `basisFormat`
        if not name.endswith("coefficients"):
            continue
        basis_fmt = dset.attrs["basisFormat"]
        dset.attrs["basisFormat"] = np.fromiter(basis_fmt[1:-1].split(","), dtype=np.int64)
    f.attrs["__version__"] = __version__


def store_cp2k_basis(path_hdf5: PathLike, path_basis: PathLike) -> None:
    """Read the CP2K basis set into an HDF5 file."""
    with h5py.File(path_hdf5, "r+") as f:
        if Version(f.attrs.get("__version__", "0.0.0")) < _NANO_QMFLOWS_0_13_0:
            _convert_legacy_basis(f)

    keys, vals = read_cp2k_basis(path_basis, allow_multiple_exponents=True)
    node_paths_exponents = [_join(xs, str(xs.exponent_set), "exponents") for xs in keys]
    node_paths_coefficients = [_join(xs, str(xs.exponent_set), "coefficients") for xs in keys]

    exponents = [xs.exponents for xs in vals]
    coefficients = [xs.coefficients for xs in vals]
    formats = [np.fromiter(xs.basisFormat, dtype=np.int64) for xs in keys]

    store_arrays_in_hdf5(path_hdf5, node_paths_exponents, exponents)
    store_arrays_in_hdf5(path_hdf5, node_paths_coefficients, coefficients,
                         attribute=BasisFormats(name="basisFormat", value=formats))


def guesses_to_compute(calculate_guesses: str, enumerate_from: int, len_geometries: int) -> List[int]:
    """Guess for the wave function."""
    if calculate_guesses is None:
        points_guess = []
    elif calculate_guesses.lower() in 'first':
        # Calculate new Guess in the first geometry
        points_guess = [enumerate_from]
        msg = "An initial Calculation will be computed as guess for the wave function"
        logger.info(msg)
    elif calculate_guesses.lower() in 'all':
        # Calculate new Guess in each geometry
        points_guess = [enumerate_from + i for i in range(len_geometries)]
        msg = "A guess calculation will be done for each geometry"
        logger.info(msg)

    return points_guess


def read_swaps(path_hdf5: Union[str, Path], project_name: str) -> Matrix:
    """Read the crossing tracking for the Molecular orbital."""
    path_swaps = join(project_name, 'swaps')
    if is_data_in_hdf5(path_hdf5, path_swaps):
        return retrieve_hdf5_data(path_hdf5, path_swaps)
    else:
        msg = f"""There is not a tracking file called: {path_swaps}
        This file is automatically created when running the worflow_coupling
        simulations"""
        raise RuntimeError(msg)


def split_trajectory(path: str | Path, nblocks: int, pathOut: str | os.PathLike[str]) -> List[str]:
    """Split an XYZ trajectory in n Block and write them in a given path.

    Parameters
    ----------
    path
        Path to the XYZ file.
    nblocks
        number of Block into which the xyz file is split.
    pathOut
        Path were the block are written.

    Returns
    -------
    list
        list of paths to the xyz geometries

    """
    with open(path, 'r') as f:
        # Read First line
        ls = f.readline()
        numat = int(ls.split()[0])

    # Number of lines in the file
    cmd = f"wc -l {path}"
    ls = subprocess.check_output(cmd.split()).decode()
    lines = int(ls.split()[0])
    if (lines % (numat + 2)) != 0:
        lines += 1

    # Number of points in the xyz file
    npoints = lines // (numat + 2)
    # Number of points for each chunk
    nchunks = int(np.ceil(npoints / nblocks))
    # Number of lines per block
    lines_per_block = nchunks * (numat + 2)
    # Path where the splitted xyz files are written
    prefix = join(pathOut, 'chunk_xyz_')
    cmd = f'split -a 1 -l {lines_per_block} {path} {prefix}'
    output = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
    rs = output.communicate()
    err = rs[1]
    if err:
        raise RuntimeError(f"Submission Errors: {err.decode()}")
    else:
        return fnmatch.filter(os.listdir(), "chunk_xyz_?")


def create_path_option(path: str | os.PathLike[str]) -> Path | None:
    """Create a Path object or return None if path is None."""
    return Path(path) if path is not None else None



"""Defaults to call CP2K.

Index
-----
.. currentmodule:: nanoqm.workflows.templates
.. autosummary::
  create_settings_from_template

"""

from __future__ import annotations

__all__ = ["create_settings_from_template"]

import os
import yaml
from scm.plams import Molecule

from qmflows import Settings
from qmflows.type_hints import PathLike
from nanoqm.common import UniqueSafeLoader, valence_electrons, aux_fit
from typing import Any, Dict, Iterable, FrozenSet


def generate_auxiliar_basis(
        sett: Settings, auxiliar_basis: str, quality: str) -> Settings:
    """Generate the `auxiliar_basis` for all the atoms in the `sett`.

    Use the`quality` of the auxiliar basis provided by the user.
    """
    quality_to_number = {"low": 0, "medium": 1,
                         "good": 2, "verygood": 3, "excellent": 4}
    kind = sett.cp2k.force_eval.subsys.kind
    for atom in kind.keys():
        index = quality_to_number[quality.lower()]
        cfit = aux_fit[atom][index]
        kind[atom]["basis_set"].append(f"AUX_FIT CFIT{cfit}")

    return sett


#: Settings for a PBE calculation to compute a guess wave function
cp2k_guess = Settings(yaml.load("""
cp2k:
  global:
    run_type:
      energy
  force_eval:
    dft:
      scf:
        eps_scf: 1e-6
        added_mos: 0
        scf_guess: "restart"
        ot:
          minimizer: "DIIS"
          n_diis: 7
          preconditioner: "FULL_SINGLE_INVERSE"
""", Loader=UniqueSafeLoader))

#: Settings for a PBE calculation to compute the Molecular orbitals
cp2k_main = Settings(yaml.load("""
cp2k:
  global:
    run_type:
      energy

  force_eval:
    dft:
      scf:
        eps_scf: 5e-4
        max_scf: 200
        scf_guess: "restart"
""", Loader=UniqueSafeLoader))


#: Settings for a PBE calculation to compute a guess wave function
cp2k_pbe_guess = Settings(yaml.load("""
cp2k:
  global:
    run_type:
      energy
  force_eval:
    subsys:
      cell:
        periodic: "None"
    dft:
      xc:
        xc_functional pbe: {}
      scf:
        eps_scf: 1e-6
        added_mos: 0
        scf_guess: "restart"
        ot:
          minimizer: "DIIS"
          n_diis: 7
          preconditioner: "FULL_SINGLE_INVERSE"
""", Loader=UniqueSafeLoader))

#: Settings for a PBE calculation to compute the Molecular orbitals
cp2k_pbe_main = Settings(yaml.load("""
cp2k:
  global:
    run_type:
      energy

  force_eval:
    subsys:
      cell:
        periodic: "None"
    dft:
      xc:
        xc_functional pbe: {}
      scf:
        eps_scf: 5e-4
        max_scf: 200
        scf_guess: "restart"
""", Loader=UniqueSafeLoader))

#: Settings for a R2SCAN calculation to compute a guess wave function
cp2k_scan_guess = Settings(yaml.load("""
cp2k:
  global:
    run_type:
      energy
  force_eval:
    subsys:
      cell:
        periodic: "None"
    dft:
      xc:
        xc_functional:
          mgga_x_r2scan:
            scale: 1.0
          mgga_c_r2scan:
            scale: 1.0
        xc_grid:
          xc_deriv: "spline3"
          xc_smooth_rho: "None"
      scf:
        eps_scf: 1e-6
        added_mos: 0
        scf_guess: "restart"
        ot:
          minimizer: "DIIS"
          n_diis: 7
          preconditioner: "FULL_SINGLE_INVERSE"
""", Loader=UniqueSafeLoader))

#: Settings for a R2SCAN calculation to compute the Molecular orbitals
cp2k_scan_main = Settings(yaml.load("""
cp2k:
  global:
    run_type:
      energy

  force_eval:
    subsys:
      cell:
        periodic: "None"
    dft:
      xc:
        xc_functional:
          mgga_x_r2scan:
            scale: 1.0
          mgga_c_r2scan:
            scale: 1.0
        xc_grid:
          xc_deriv: "spline3"
          xc_smooth_rho: "None"
      scf:
        eps_scf: 5e-4
        max_scf: 200
        scf_guess: "restart"
""", Loader=UniqueSafeLoader))

#: Settings for a PBE0 calculation to compute a guess wave function
cp2k_pbe0_guess = Settings(yaml.load("""
cp2k:
   global:
     run_type:
       energy

   force_eval:
     subsys:
       cell:
         periodic: "None"
     dft:
       auxiliary_density_matrix_method:
         method: "basis_projection"
         admm_purification_method: "none"
       qs:
         method: "gpw"
         eps_pgf_orb: 1E-8
       xc:
        xc_functional:
          pbe:
            scale_x: 0.75
            scale_c: 1.00
        hf:
          fraction: 0.25
          screening:
            eps_schwarz: 1.0E-6
            screen_on_initial_p: "True"
          interaction_potential:
            potential_type: "truncated"
            cutoff_radius: 2.5
          memory:
            max_memory: 5000
            eps_storage_scaling: "0.1"
       scf:
          eps_scf: 1e-6
          added_mos: 0
          scf_guess: "restart"
          ot:
            minimizer: "DIIS"
            n_diis: 7
            preconditioner: "FULL_SINGLE_INVERSE"

""", Loader=UniqueSafeLoader))

#: Settings for a PBE0 calculation to compute the Molecular orbitals
cp2k_pbe0_main = Settings(yaml.load("""
cp2k:
   global:
     run_type:
       energy

   force_eval:
     subsys:
       cell:
         periodic: "None"
     dft:
       auxiliary_density_matrix_method:
         method: "basis_projection"
         admm_purification_method: "none"
       qs:
         method: "gpw"
         eps_pgf_orb: "1.0E-8"
       xc:
         xc_functional:
           pbe:
             scale_x: "0.75"
             scale_c: "1.00"
         hf:
           fraction: "0.25"
           screening:
             eps_schwarz: 1.0E-6
             screen_on_initial_p: "True"
           interaction_potential:
             potential_type: "truncated"
             cutoff_radius: 2.5
           memory:
             max_memory: "5000"
             eps_storage_scaling: "0.1"
       scf:
          eps_scf: 5e-4
          max_scf: 200
          scf_guess: "restart"
""", Loader=UniqueSafeLoader))

#: Settings for a HSE06 calculation to compute a guess wave function
cp2k_hse06_guess = Settings(yaml.load("""
cp2k:
   global:
     run_type:
       energy

   force_eval:
     subsys:
       cell:
         periodic: "None"
     dft:
       auxiliary_density_matrix_method:
         method: "basis_projection"
         admm_purification_method: "none"
       qs:
         method: "gpw"
         eps_pgf_orb: 1E-8
       xc:
        xc_functional:
          pbe:
            scale_x: 0.00
            scale_c: 1.00
          xwpbe:
            scale_x: -0.25
            scale_x0: 1.00
            omega: 0.11
        hf:
          fraction: 0.25
          screening:
            eps_schwarz: 1.0E-6
            screen_on_initial_p: "True"
          interaction_potential:
            potential_type: "shortrange"
            omega: 0.11
          memory:
            max_memory: 5000
            eps_storage_scaling: "0.1"
       scf:
          eps_scf: 1e-6
          added_mos: 0
          scf_guess: "restart"
          ot:
            minimizer: "DIIS"
            n_diis: 7
            preconditioner: "FULL_SINGLE_INVERSE"

""", Loader=UniqueSafeLoader))

#: Settings for a HSE06 calculation to compute the Molecular orbitals
cp2k_hse06_main = Settings(yaml.load("""
cp2k:
   global:
     run_type:
       energy

   force_eval:
     subsys:
       cell:
         periodic: "None"
     dft:
       auxiliary_density_matrix_method:
         method: "basis_projection"
         admm_purification_method: "none"
       qs:
         method: "gpw"
         eps_pgf_orb: "1.0E-8"
       xc:
        xc_functional:
          pbe:
            scale_x: 0.00
            scale_c: 1.00
          xwpbe:
            scale_x: -0.25
            scale_x0: 1.00
            omega: 0.11
        hf:
          fraction: 0.25
          screening:
            eps_schwarz: 1.0E-6
            screen_on_initial_p: "True"
          interaction_potential:
            potential_type: "shortrange"
            omega: 0.11
          memory:
            max_memory: 5000
            eps_storage_scaling: "0.1"
       scf:
          eps_scf: 1e-6
          max_scf: 200
          scf_guess: "restart"
""", Loader=UniqueSafeLoader))

#: Settings for a B3LYP calculation to compute a guess wave function
cp2k_b3lyp_guess = Settings(yaml.load("""
cp2k:
   global:
      run_type:
         energy

   force_eval:
      subsys:
         cell:
           periodic: "None"
      dft:
         xc:
           xc_functional b3lyp: {}
         scf:
           eps_scf: 1e-6
           added_mos: 0
           scf_guess: "restart"
           ot:
             minimizer: "DIIS"
             n_diis: 7
             preconditioner: "FULL_SINGLE_INVERSE"
""", Loader=UniqueSafeLoader))

#: Settings for a B3LYP calculation to compute the Molecular orbitals
cp2k_b3lyp_main = Settings(yaml.load("""
cp2k:
  global:
    run_type:
      energy

  force_eval:
    subsys:
      cell:
        periodic: "None"
    dft:
      xc:
        xc_functional b3lyp: {}
      scf:
        eps_scf: 5e-4
        max_scf: 200
        scf_guess: "restart"
""", Loader=UniqueSafeLoader))


#: Settings to add the CP2K kinds for each atom
kinds_template = Settings(yaml.load("""
cp2k:
  force_eval:
    subsys:
      kind:
        C:
          basis_set: DZVP-MOLOPT-SR-GTH-q4
          potential: GTH-PBE-q4
""", Loader=UniqueSafeLoader))


def generate_kinds(elements: Iterable[str], basis: str, potential: str) -> Settings:
    """Generate the kind section for cp2k basis."""
    s = Settings()
    subsys = s.cp2k.force_eval.subsys
    for e in elements:
        q = valence_electrons[e]
        subsys.kind[e]['basis_set'] = [f"{basis}-q{q}"]
        subsys.kind[e]['potential'] = f"{potential}-q{q}"

    return s


#: available templates
templates_dict = {
    "guess": cp2k_guess, "main": cp2k_main,
    "pbe_guess": cp2k_pbe_guess, "pbe_main": cp2k_pbe_main,
    "scan_guess": cp2k_scan_guess, "scan_main": cp2k_scan_main,
    "pbe0_guess": cp2k_pbe0_guess, "pbe0_main": cp2k_pbe0_main,
    "hse06_guess": cp2k_hse06_guess, "hse06_main": cp2k_hse06_main,
    "b3lyp_guess": cp2k_b3lyp_guess, "b3lyp_main": cp2k_b3lyp_main}


def create_settings_from_template(
    general: Dict[str, Any],
    template_name: str,
    path_traj_xyz: str | os.PathLike[str],
) -> Settings:
    """Create a job Settings using the name provided by the user."""
    setts = templates_dict[template_name]
    elements = read_unique_atomic_labels(path_traj_xyz)

    kinds = generate_kinds(elements, general['basis'], general['potential'])

    if 'pbe0' in template_name:
        s = Settings()
        return generate_auxiliar_basis(setts + s + kinds, general['basis'], general['aux_fit'])
    elif 'hse06' in template_name:
        return generate_auxiliar_basis(setts + kinds, general['basis'], general['aux_fit'])
    else:
        return setts + kinds


def read_unique_atomic_labels(path_traj_xyz: str | os.PathLike[str]) -> FrozenSet[str]:
    """Return the unique atomic labels."""
    mol = Molecule(path_traj_xyz, 'xyz')

    return frozenset(at.symbol for at in mol.atoms)
